/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import bsh.EvalError;
import bsh.Interpreter;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.listener.EventListenerExecution;
import org.jbpm.pvm.model.Condition;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.ExpressionEvaluationException;
import org.ow2.bonita.runtime.var.Enumeration;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.novabpm.util.Misc;


/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class ConditionDefinition implements Condition {
  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(ConditionDefinition.class.getName());
  /**
   * contains the condition set onto the transition; for instance:
   * "!str1.equals("toto")" where str1 is the name of a runtime dataField
   */
  protected String value;
  protected String scriptType;

  protected ConditionDefinition() {
  }

  public ConditionDefinition(String value, String scriptType) {
    this.value = value;
    this.scriptType = scriptType;
  }

  public boolean evaluate(EventListenerExecution xpdlExecution) {
    if ("bsh".equals(scriptType)) {
      return evaluateBsh(xpdlExecution);
    }
    throw new BonitaRuntimeException("Unsupported scriptType : " + scriptType);
  }

  private boolean evaluateBsh(EventListenerExecution xpdlExecution) {
    Set<String> nullVariables = new HashSet<String>();
    try {
      boolean bool = false;
      Interpreter i = new Interpreter();
      // to set all process variables to the interpreter
      Set<String> setVar = xpdlExecution.getVariableKeys();
      Iterator<String> it = setVar.iterator();
      
      while (it.hasNext()) {
        String key = it.next();
        Object var = xpdlExecution.getVariable(key);
        if (var == null) {
          nullVariables.add(key);
          i.set(key, null);
        } else if (var instanceof String) {
          i.set(key, var);
        } else if (var instanceof Enumeration) {
          Enumeration enu = (Enumeration) var;
          i.set(key, enu.getSelectedValue());
        }
      }
      // set result with the value evaluation
      i.eval("result = " + value);
      if (i.get("result") != null) {
        bool = ((Boolean) i.get("result")).booleanValue();
        return bool;
      }
      throw new BonitaWrapperException(
          new ExpressionEvaluationException("Evaluation of the condition by the interpreter is null !"));
    } catch (EvalError e) {
      String end = "The following variables were null when evaluating the condition, "
        + "maybe one of those is used in the expression : ";
      for (String nullVariable : nullVariables) {
        end += "\n- " + nullVariable;
      }
      //bsh.CallStack is not Serializable
      String stacktrace = Misc.getStackTraceFrom(e);
      String errorMsg = "Error during evaluation of condition : " + value + ". " + end + "\n " + stacktrace;
      if (LOG.isLoggable(Level.INFO)) {
        LOG.severe(errorMsg);
      }
      
      throw new BonitaWrapperException(new ExpressionEvaluationException(errorMsg));
    }
  }

  public String getValue() {
    return value;
  }

}
