/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.Hook;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.TxHook;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.StandaloneAccessorUtil;


public final class HookExecutor {

  private static Logger log = Logger.getLogger(HookExecutor.class.getName());

  private HookExecutor() { }

  public static void executeHook(XpdlExecution execution, String activityId, JavaHook javaHook) {
    if (javaHook != null) {
      final XpdlInstance instance = execution.getXpdlInstance();
      final ProcessInstanceUUID instanceUUID = instance.getUUID();
      final String iterationId = execution.getIterationId();
      final String activityInstanceId = execution.getActivityInstanceId();
      final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
      boolean inTransaction  = javaHook.isInTransaction();
      try {
        Object hookInstance = ClassDataLoader.getHookInstance(instance.getPackageDefinitionUUID(), javaHook);
        if (!inTransaction && hookInstance instanceof TxHook) {
          throw new BonitaRuntimeException("The specified hook eventName declares an 'outside transaction hook' "
            + "but the given class implements " + TxHook.class.getName()  + " instead of " + Hook.class.getName());
        }
        if (log.isLoggable(Level.INFO)) {
          log.info("Starting hook (instance=" + instanceUUID + ", process=" + processUUID 
              + ", activityId=" + activityId + ") : " + javaHook);
        }
        ActivityInstance<ActivityBody> activityInst = 
          EnvTool.getJournalQueriers().getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId);
        if (hookInstance instanceof Hook) {
          Hook hook = (Hook) hookInstance;
          hook.execute(StandaloneAccessorUtil.getQueryAPIAccessor(), activityInst);
        } else if (hookInstance instanceof TxHook) {
          TxHook txHook = (TxHook) hookInstance;
          txHook.execute(StandaloneAccessorUtil.getAPIAccessor(), activityInst);
        } else {
          throw new BonitaRuntimeException("Specified class is not a valid hook. It does not implement "
            + Hook.class.getName() + " neither " + TxHook.class.getName());
        }
        if (log.isLoggable(Level.INFO)) {
          log.info("Finished hook (instance=" + instanceUUID + ", process=" + processUUID 
              + ", activityId=" + activityId + ") : " + javaHook);
        }
      } catch (Exception e) {
        if (inTransaction) {
          throw new BonitaRuntimeException("Exception caught while executing hook : " + javaHook, e);
        } else if (log.isLoggable(Level.SEVERE)) {
          log.severe("Exception caught while executing hook (instance=" + instanceUUID + ", process=" + processUUID 
              + ", activityId=" + activityId + ") : " + javaHook + " - Exception : " + e);
          e.printStackTrace();
        }
      }
    }
  }

  public static void executeHooks(List<JavaHook> javaHooks, XpdlExecution execution, String activityId, JavaHook.Type type) {
    if (javaHooks != null) {
      for (JavaHook javaHook : javaHooks) {
        if (javaHook.getType().equals(type)) {
          executeHook(execution, activityId, javaHook);
        }
      }
    }
  }
}


