/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.activity.ActivityExecution;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.impl.SubflowBodyImpl;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.SubflowBodyUUID;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class SubFlow extends AbstractActivity {

  protected static final Logger LOG = Logger.getLogger(SubFlow.class.getName());

  public static final String SUBFLOW_SIGNAL = "end_of_subflow";
  protected String subProcessId;
  protected boolean isAsync;
  // MAP: key = parameter name, value = name of the variable that contains the value.
  protected Map<String, String> inParameters;
  protected Map<String, String> outParameters;

  protected  SubFlow() {
    super();
  }

  public SubFlow(String activityName, JoinType joinType,
      SplitType splitType,
      Performer performer,
      String subProcessId,
      boolean isAsync,
      Map<String,String> inParameters,
      Map<String,String> outParameters) {

    super(activityName, joinType,
        splitType,
        ActivityType.subFlow,
        performer
        );
    this.subProcessId = subProcessId;
    this.isAsync = isAsync;
    this.inParameters = inParameters;
    this.outParameters = outParameters;
  }

  @Override
  public ActivityBody getBody(XpdlExecution xpdlExecution, ActivityInstance<ActivityBody> activityInstance) {
    return new SubflowBodyImpl(new SubflowBodyUUID(activityInstance.getUUID()));
  }

  @Override
  protected boolean bodyStartAutomatically() {
    return true;
  }

  @Override
  protected boolean executeBusinessLogic(Execution execution) {
    Repository repository = EnvTool.getRepository();
    final XpdlProcess process = repository.findLatestProcessById(subProcessId);
    if (process == null) {
      throw new BonitaRuntimeException("SubFlow : process not found " + subProcessId);
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.finest("Starting a new instance of process : " + subProcessId);
    }
    Map<String, Object> parameterValues = new HashMap<String, Object>();
    if (inParameters != null) {
      for (Map.Entry<String, String> parameter : inParameters.entrySet()) {
        parameterValues.put(parameter.getKey(), ((XpdlExecution)execution).getVariable(parameter.getValue()));
      }
    }
    final XpdlExecution rootExecution = (XpdlExecution) process.beginProcessInstance();
    rootExecution.setVariables(parameterValues);
    final XpdlInstance instance = rootExecution.getXpdlInstance();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.finest("Started subprocess instance : " + instance);
    }
    if (!isAsync) {
      instance.setExecutionToSignal((XpdlExecution)execution);
    }
    rootExecution.getExecution(process.getInitial().getName()).signal();
    if (isAsync) {
      return true;
    }
    return false;
  }

  @Override
  public void signal(ActivityExecution execution, String signal, Map<String, Object> signalParameters) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    if (SUBFLOW_SIGNAL.equals(signal)) {
      if (outParameters != null) {
        final Recorder recorder = EnvTool.getRecorder();
        ProcessInstanceUUID instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
        String activityId = xpdlExecution.getNode().getName();

        for (Map.Entry<String, String> parameter : outParameters.entrySet()) {
          String variableId = parameter.getValue();
          Object variableValue = signalParameters.get(parameter.getKey());
          execution.setVariable(variableId, variableValue);

          if (xpdlExecution.getXpdlInstance().getRootExecution().hasVariable(variableId)) {
            recorder.recordInstanceVariableUpdated(variableId, variableValue, instanceUUID, EnvTool.getUserId());
          } else {
            recorder.recordActivityVariableUpdated(variableId, variableValue, instanceUUID, activityId,
                xpdlExecution.getIterationId(), xpdlExecution.getActivityInstanceId(), EnvTool.getUserId());
          }
        }
      }
      super.signal(xpdlExecution, BODY_FINISHED, null);
    } else {
      super.signal(xpdlExecution, signal, signalParameters);
    }
  }

}


