/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.deployment;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.jbpm.pvm.internal.model.NodeImpl;
import org.jbpm.pvm.internal.model.ProcessDefinitionImpl;
import org.jbpm.pvm.internal.model.TransitionImpl;
import org.jbpm.pvm.model.Node;
import org.jbpm.pvm.model.Transition;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.definition.activity.AbstractActivity.IterationDescriptor;
import org.ow2.bonita.definition.activity.AbstractActivity.JoinType;
import org.ow2.bonita.definition.activity.AbstractActivity.SplitType;
import org.ow2.novabpm.util.Misc;


/**
 * @author Guillaume Porcher
 *
 */
public final class IterationDetection {
  private static final Logger LOG = Logger.getLogger(IterationDetection.class.getName());

  /**
   * Helper object to store data about cycle structure
   *
   * @author Guillaume Porcher
   */
  private static class CycleObject {
    // List of entry nodes
    protected Set<String> entryNodes = new HashSet<String>();
    // List of exit nodes
    protected Set<String> exitNodes = new HashSet<String>();
    // List of nodes in the cycle
    protected Set<String> path = new HashSet<String>();
    // List of transitions that are in the cycle
    protected Set<MyTransition> transitions = new HashSet<MyTransition>();

    @Override
    public String toString() {
      return "entryNodes: " + entryNodes + ", exitNodes: " + exitNodes + " path: " + path + ", transitions: " + transitions;
    }
  }

  /**
   * Simple object that describes a transition.
   *
   * @author Guillaume Porcher
   */
  private static class MyTransition {
    // source node
    protected String fromNode;
    // destination node
    protected String toNode;
    @Override
    public String toString() {
      return fromNode + "->" + toNode;
    }
  }

  private IterationDetection() { }

  /**
   * Creates a spanning tree in spanningTree corresponding to the input process
   * and lists transitions that are not in the spanning tree.
   * Use a Deep First Search Algorithm for that.
   * Transitions that are not in the spanning tree can be:
   * - transitions that creates a second path between two nodes (e.g. split/join)
   * - transitions that creates loops
   * @param inProcess input process
   * @param spanningTree process where the corresponding spanning tree will be created
   * @return list of transitions of input process that are not in the spanning tree.
   */
  private static List<Transition> createSpanningTree(XpdlProcess inProcess, ProcessDefinitionImpl spanningTree) {
    Node inNode = inProcess.getInitial();
    NodeImpl outNode = spanningTree.createNode(inNode.getName());

    // Create a spanning tree in outProcess and list transitions that are not in the spanning tree.
    // Use a Deep First Search Algorithm for that.
    return processNode(inNode, outNode, spanningTree);
  }

  /**
   * Deep first search algorithm that creates spanning tree
   * (see {@link IterationDetection#createSpanningTree(ProcessDefinitionImpl, ProcessDefinitionImpl)}
   * @param inNode node of the input process currently processed
   * @param outNode corresponding node in the spanning tree
   * @param outProcess spanning tree
   * @return list of unprocessed transitions during the DFS
   */
  private static List<Transition> processNode(Node inNode, NodeImpl outNode, ProcessDefinitionImpl outProcess) {
    List<Transition> nonProcessedTransitions = new ArrayList<Transition>();
    if (inNode.hasOutgoingTransitions()) {
      for (Transition t : inNode.getOutgoingTransitions()) {
        NodeImpl destNode = (NodeImpl) t.getDestination();
        String destName = destNode.getName();
        if (!outProcess.hasNode(destName)) {
          // create node
          NodeImpl newNode = outProcess.createNode(destName);
          outNode.createOutgoingTransition(newNode, t.getName());
          nonProcessedTransitions.addAll(processNode(t.getDestination(), newNode, outProcess));
        } else {
          // node already in the spanning tree, the transition is not in the spanning tree.
          nonProcessedTransitions.add(t);
        }
      }
    }
    return nonProcessedTransitions;
  }

  /**
   * Checks if there is a path from source to destination
   * @param sourceNode source
   * @param destNode destination
   * @return true if there is a path from source to destination
   */
  private static boolean existsPath(NodeImpl sourceNode, NodeImpl destNode) {
    if (sourceNode.equals(destNode)) {
      return true;
    }
    if (sourceNode.hasOutgoingTransitions()) {
      for (Transition t : sourceNode.getOutgoingTransitions()) {
        if (existsPath((NodeImpl) t.getDestination(), destNode)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Find transitions in the given list of transitions that would form a cycle if added to the given process.
   * If a transition can be added to the process without creating a cycle, then the transition is added.
   * The process returned is a DAG (directed acyclic graph)
   * @param process input process (this process will be modified by calling this method)
   * @param transitions transitions to check
   * @return list of transitions that cannot be added to the DAG (transitions that would create a cycle).
   */
  private static List<Transition> findCycleTransitions(ProcessDefinitionImpl process, List<Transition> transitions) {
    List<Transition> cycleTransitions = new ArrayList<Transition>();
    for (Transition t : transitions) {
      String sourceNodeName = t.getSource().getName();
      String destNodeName = t.getDestination().getName();

      // Check if when a transition is added to the spanning tree, it creates a cycle
      // check if there is already a path from transition dest to transition source.
      // adding the transition would then create a cycle.
      NodeImpl sourceNode = process.getNode(sourceNodeName);
      NodeImpl destNode = process.getNode(destNodeName);
      if (existsPath(destNode, sourceNode)) {
        cycleTransitions.add(t);
      } else {
        // add transition to the DAG
        sourceNode.createOutgoingTransition(destNode);
      }
    }
    return cycleTransitions;
  }


  /**
   * Fill the path and transitions fields of the CycleObject c.
   * @param c cycle object to fill
   * @param sourceNode source node of the path
   * @param destNode destination of the path
   * @return
   */
  private static boolean fillCycleObject(CycleObject c, NodeImpl sourceNode, NodeImpl destNode) {
    if (sourceNode.equals(destNode)) {
      c.path.add(sourceNode.getName());
      return true;
    }
    if (sourceNode.hasOutgoingTransitions()) {
      boolean res = false;
      for (Transition t : sourceNode.getOutgoingTransitions()) {
        if (fillCycleObject(c, (NodeImpl) t.getDestination(), destNode)) {
          c.path.add(sourceNode.getName());
          MyTransition myT = new MyTransition();
          myT.fromNode = t.getSource().getName();
          myT.toNode = t.getDestination().getName();
          c.transitions.add(myT);
          res = true;
        }
      }
      return res;
    }
    return false;
  }

  /**
   * Find entry nodes in the cycle. Checks that entry nodes does not use Xor Join.
   * @param c cycle object to fill
   * @param inProcess input process (xpdl process)
   */
  private static void fillCycleObjectEntryNodes(CycleObject c, XpdlProcess inProcess) {
    boolean hasEntryPointXor = false;
    for (String nodeName : c.path) {
      NodeImpl sourceNode = inProcess.getNode(nodeName);
      if (sourceNode.hasIncomingTransitions()) {
        // check for entryNode: 
        // n is an entry node if: 
        // - a transition from a node that is not in the cycle goes to n.
        // - a transition which is not in the cycle goes to n.
        for (Transition t : sourceNode.getIncomingTransitions()) {
          JoinType joinType;
          boolean isEntryNode = true;
          if (c.path.contains(t.getSource().getName())) {
            for (MyTransition myTr : c.transitions) {
              if (myTr.fromNode.equals(t.getSource().getName()) 
                  && myTr.toNode.equals(t.getDestination().getName())) {
                isEntryNode = false;
              }
            }
          }
          if (isEntryNode) {
            c.entryNodes.add(sourceNode.getName());
            joinType = ((AbstractActivity)sourceNode.getBehaviour()).getJoinType(); 
            if (JoinType.XOR.equals(joinType)) {
              hasEntryPointXor = true; 
            }
          }
        }
      }
    }
    if (c.entryNodes.size() == 0) {
      throw new DeploymentRuntimeException("Error in cycle detection : cycle " + c + " has no start node");
    }
    if (!hasEntryPointXor) {
      throw new DeploymentRuntimeException("Error in cycle detection : cycle " + c + " has no start node with a XOR join." 
          + " Process execution can never enter this cycle.");
    }
  }

  /**
  * Find exit nodes in the cycle. Checks if exit nodes only use Xor split type.
  * @param c cycle object to fill
  * @param inProcess input process (xpdl process)
  */
  private static void checkExitNodes(CycleObject c, XpdlProcess inProcess) {
    for (String nodeName : c.path) {
      NodeImpl sourceNode = inProcess.getNode(nodeName);
      if (sourceNode.hasOutgoingTransitions()) {
        //check for exit
        for (Transition t : sourceNode.getOutgoingTransitions()) {
          if (!c.path.contains(t.getDestination().getName())) {
            c.exitNodes.add(sourceNode.getName());
            // sourceNode is an exit node
            SplitType splitType = ((AbstractActivity)sourceNode.getBehaviour()).getSplitType(); 
            // Only allow XOR split for exit nodes
            if (!SplitType.XOR.equals(splitType)) {
              // TODO: change this to an exception
//              throw new BonitaRuntimeException(
//                  "Error in iteration : " + nodeName + " is an exit node for cycle " + c.path + "." + Misc.LINE_SEPARATOR
//                  + "Split type of this node is " + splitType + " but only XOR is allowed.");
              LOG.severe("Potential issue in iteration : " + nodeName + " is an exit node for cycle " 
                  + c.path + "." + Misc.LINE_SEPARATOR
                  + "Split type of this node is " + splitType + " but only XOR is supported." + Misc.LINE_SEPARATOR
                  + "Runtime behaviour is undefined if more than one transition is enabled at the same time.");
            }
          }
        }
      }
    }
  }
  
  /**
   * Create structure to add to the XpdlProcess to handle cycles.
   * For that, the cycle structure is checked, then entry nodes are resolved.
   * @param inProcess XpdlProcess
   * @param dag acyclic graph on which cycle structure will be found
   * @param transitions list of transitions that creates cycles
   * @return a map where the key is the name of a node in the input process,
   * and the value a list of IterationDescriptor that needs to be added to the node definition.
   */
  private static Map<String, List<IterationDescriptor>> findCycleInits(XpdlProcess inProcess,
      ProcessDefinitionImpl dag, List<Transition> transitions) {

    Map<String, List<IterationDescriptor>> iterationDescriptors = new HashMap<String, List<IterationDescriptor>>();

    for (Transition t : transitions) {
      String sourceNodeName = t.getSource().getName();
      String destNodeName = t.getDestination().getName();

      NodeImpl sourceNode = dag.getNode(sourceNodeName);
      NodeImpl destNode = dag.getNode(destNodeName);


      // Add current transition to the DAG to create a cycle
      TransitionImpl transition  = (TransitionImpl) sourceNode.createOutgoingTransition(destNode);

      // Create object that represents the cycle structure
      CycleObject c = new CycleObject();
      // compute the cycle path and transitions
      if (fillCycleObject(c, destNode, sourceNode)) {
        MyTransition myT = new MyTransition();
        myT.fromNode = sourceNodeName;
        myT.toNode = destNodeName;
        c.transitions.add(myT);
      }
      // find entry nodes
      fillCycleObjectEntryNodes(c, inProcess);

      // check exit nodes
      checkExitNodes(c, inProcess);
      
      LOG.info("CycleObject :"  + c);

      // Remove added transition to have a DAG for other transitions.
      sourceNode.removeOutgoingTransition(transition);
      destNode.removeIncomingTransition(transition);

      // Create IterationDescriptors for the XpdlProcess
      for (String startNode : c.entryNodes) {
        if (!iterationDescriptors.containsKey(startNode)) {
          iterationDescriptors.put(startNode, new ArrayList<IterationDescriptor>());
        }
        for (String joinNode : c.path) {
          Set<String> transitionSourceNodes = new HashSet<String>();
          for (MyTransition mt : c.transitions) {
            if (mt.toNode.equals(joinNode)) {
              transitionSourceNodes.add(mt.fromNode);
            }
          }
          IterationDescriptor itDescr = new IterationDescriptor(joinNode, transitionSourceNodes);
          iterationDescriptors.get(startNode).add(itDescr);
        }
      }
    }
    return iterationDescriptors;
  }
  
  /**
   * Find cycles in the xpdlProcess and add the required structures to handle iterations
   * @param inProcess process to check and modify
   * @return inProcess
   */
  public static XpdlProcess findIterations(XpdlProcess inProcess) {
    ProcessDefinitionImpl outProcess = new ProcessDefinitionImpl();
    // Create a spanning tree in outProcess and list transitions that are not in the spanning tree.
    // Use a Deep First Search Algorithm for that.
    List<Transition> unprocessedTransitions = createSpanningTree(inProcess, outProcess);
    LOG.info("Unprocessed transitions: " + unprocessedTransitions);
    // Transitions can be out of the spanning tree because:
    // it's in a split join
    // or it's a cycle
    // The second pass determines if the transition creates a cycle and creates a DAG (Directed acyclic graph)
    List<Transition> cycleTransitions = findCycleTransitions(outProcess, unprocessedTransitions);
    LOG.info("Cycle transitions: " + cycleTransitions);

    // create Iteration descriptors
    Map<String,List<IterationDescriptor>> iterationDescriptors = findCycleInits(inProcess, outProcess, cycleTransitions);
    LOG.info("iteration descriptors: " + iterationDescriptors);

    // update XpdlProcess
    for (Map.Entry<String, List<IterationDescriptor>> itDesc : iterationDescriptors.entrySet()) {
      AbstractActivity activity = (AbstractActivity) inProcess.getNode(itDesc.getKey()).getBehaviour();
      activity.setIterationDescriptors(itDesc.getValue());
    }
    
    return inProcess;
  }
}
