/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.env;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.env.EnvironmentFactory;
import org.ow2.novabpm.util.Misc;


public final class GlobalEnvironmentFactory {

  private GlobalEnvironmentFactory() { }

  public static final String DEFAULT_ENVIRONMENT = "environment.xml";

  private static final Logger LOG = Logger.getLogger(GlobalEnvironmentFactory.class.getName());
  private static EnvironmentFactory environmentFactory;
  /**
   * Property key used to define the environment XML file to parse for the
   * holding of the {@link EnvironmentFactory} and returned by {@link #getEnvironmentFactory()}.
   *
   * This key is set to: <code>org.ow2.bonita.environment</code>
   */
  public static final String ENVIRONMENT_PROPERTY = "org.ow2.bonita.environment";
  private static final Object LOCK = new Object();

  static class BonitaShutdownHook extends Thread {
    public BonitaShutdownHook() {
      super();
    }
    public void run() {
      try {
        if (environmentFactory != null) {
          environmentFactory.close();
          environmentFactory = null;
        }
      } catch (Exception ee) {
        ee.printStackTrace();
      }
    }
  }


  public static EnvironmentFactory getEnvironmentFactory() {
    synchronized (LOCK) {
      if (environmentFactory == null) {
        final String environmentResource = System.getProperty(ENVIRONMENT_PROPERTY, DEFAULT_ENVIRONMENT);
        // search in this order: URL, FILE, RESOURCE
        URL url = null;
        // check if url is a valid url.
        // a valid url must contain the protocol.
        try {
          url = new URL(environmentResource);
        } catch (MalformedURLException e) {
          // url is not valid, try from file
          final File file = new File(environmentResource);
          if (file.exists()) {
            try {
              // Call toURI.toURL because toURL does not escape special characters.
              // see File.toURL() javadoc.
              url = file.toURI().toURL();
            } catch (MalformedURLException e1) {
              Misc.unreachableStatement();
            }
          }
        }
        // if environmentResource is not a url, not a file
        // search in resources. 
        if (url == null) {
          url = Thread.currentThread().getContextClassLoader().getResource(environmentResource);
        }
        if (url == null) {
          throw new EnvironmentNotFoundException(environmentResource);
        }
        if (LOG.isLoggable(Level.CONFIG)) {
          LOG.config("Reading environment configuration from: " + url);
        }
        final byte[] content;
        try {
          content = Misc.getAllContentFrom(url);
        } catch (IOException e) {
          throw new InvalidEnvironmentException("An IO error occurs while reading environment from: " + url
              + " Do you have read access to this ressource?", e);
        }
        final String environment = new String(content);
        if (LOG.isLoggable(Level.CONFIG)) {
          LOG.config("The environment resource " + url + " contains: " + Misc.LINE_SEPARATOR
              + environment);
        }
        final EnvironmentFactory factory = XpdlEnvironmentParser.parseEnvironmentFactoryFromXmlString(environment);
        setEnvironmentFactory(factory);
        Runtime.getRuntime().addShutdownHook(new BonitaShutdownHook());
      }
      if (environmentFactory == null) {
        throw new InvalidEnvironmentException("environmentFactory is null.");
      }
      return environmentFactory;
    }
  }

  public static void setEnvironmentFactory(final EnvironmentFactory envFactory) {
    synchronized (LOCK) {
      environmentFactory = envFactory;
    }
  }
  
  public static boolean isInitialized() {
    return environmentFactory != null;
  }
}


