/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.env;

import java.io.InputStream;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilderFactory;

import org.jbpm.pvm.env.EnvironmentFactory;
import org.jbpm.pvm.env.PvmEnvironmentFactory;
import org.jbpm.pvm.internal.env.Authentication;
import org.jbpm.pvm.internal.env.PvmEnvironmentFactoryParser;
import org.jbpm.pvm.internal.wire.descriptor.ObjectDescriptor;
import org.jbpm.pvm.internal.xml.Bindings;
import org.jbpm.pvm.internal.xml.Parse;
import org.jbpm.pvm.internal.xml.Problem;
import org.ow2.bonita.env.binding.APIBinding;
import org.ow2.bonita.env.binding.ChainerBinding;
import org.ow2.bonita.env.binding.HibernateXpdlPersistenceServiceBinding;
import org.ow2.bonita.env.binding.ImplementationBinding;
import org.ow2.bonita.env.binding.QuerierListAccessorBinding;
import org.ow2.novabpm.util.Misc;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
  * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class XpdlEnvironmentParser extends PvmEnvironmentFactoryParser {

  public static final String CLASSPATH_URL = "classpath://";
  private static Logger log = Logger.getLogger(XpdlEnvironmentParser.class.getName());

  private static PvmEnvironmentFactoryParser instance;

  public static synchronized PvmEnvironmentFactoryParser getInstance() {
    if (instance == null) {
      instance = new XpdlEnvironmentParser();
      // Get bindings
      Bindings bindings = instance.getEnvironmentXmlParser().getBindings();
      // Use the same bindings in block and application contexts
      instance.getEnvironmentFactoryXmlParser().setBindings(bindings);
      // Add bonita bindings
      bindings.addBinding(new HibernateXpdlPersistenceServiceBinding());
      bindings.addBinding(new ImplementationBinding("repository"));
      bindings.addBinding(new ImplementationBinding("finished-instance-handler"));
      bindings.addBinding(new ImplementationBinding("undeployed-package-handler"));
      bindings.addBinding(new ImplementationBinding("recorder"));
      bindings.addBinding(new ImplementationBinding("journal"));
      bindings.addBinding(new ChainerBinding());
      bindings.addBinding(new ImplementationBinding("history"));
      bindings.addBinding(new ImplementationBinding("archiver"));
      bindings.addBinding(new QuerierListAccessorBinding());
      bindings.addBinding(new APIBinding());

    }
    return instance;
  }

  /* (non-Javadoc)
   * @see org.jbpm.env.xml.EnvironmentParser#parseDocument(org.w3c.dom.Document, org.jbpm.xml.Parse)
   */
  @Override
  public PvmEnvironmentFactory parseDocument(Document document, Parse parse) {
    // Default parsing
    PvmEnvironmentFactory defaultEnvironmentFactory = (PvmEnvironmentFactory) super.parseDocument(document, parse);

    // Add authentication descriptor to application block
    ObjectDescriptor authenticationDescriptor = new ObjectDescriptor();
    authenticationDescriptor.setClassName(Authentication.class.getName());
    defaultEnvironmentFactory.getEnvironmentFactoryCtxWireContext().getWireDefinition().addDescriptor(authenticationDescriptor);

    return defaultEnvironmentFactory;
  }

  public static EnvironmentFactory parseEnvironmentFactoryFromXmlString(
      String xmlString) {
    Parse parse = getInstance().createParse();
    parse.setString(xmlString);
    EnvironmentFactory factory = (EnvironmentFactory) parse.execute().getDocumentObject();
    parse.checkProblems("environment");
    return factory;
  }

  public synchronized DocumentBuilderFactory getDocumentBuilderFactory() {
    documentBuilderFactory = newDocumentBuilderFactory();
    documentBuilderFactory.setNamespaceAware(true);
    documentBuilderFactory.setValidating(true);
    // ignore white space can only be set if parser is validating
    documentBuilderFactory.setIgnoringElementContentWhitespace(true);
    // select xml schema as the schema language (a.o.t. DTD)
    documentBuilderFactory.setAttribute(
        "http://java.sun.com/xml/jaxp/properties/schemaLanguage",
        "http://www.w3.org/2001/XMLSchema");
    //  set schema sources
    URL url = Thread.currentThread().getContextClassLoader().getResource("environment-bonita.xsd");
    documentBuilderFactory.setAttribute(
        "http://java.sun.com/xml/jaxp/properties/schemaSource", url.toExternalForm());
    return documentBuilderFactory;
  }

  /* (non-Javadoc)
   * @see org.jbpm.xml.Parser#resolveEntity(java.lang.String, java.lang.String)
   */
  @Override
  public InputSource resolveEntity(String publicId, String systemId) {
    if (systemId.startsWith(CLASSPATH_URL)) {
      String localpart = systemId.substring(CLASSPATH_URL.length());
      InputStream inStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(localpart);
      if (inStream != null) {
        return new InputSource(inStream);
      }
    }
    return super.resolveEntity(publicId, systemId);
  }

  /** throws an exception with appropriate message in case the parse contains
   * errors or fatal errors.  This method also logs the problems with severity
   * 'warning'. */
  public void checkProblems(String description, Parse parse) {
    if (parse.hasProblems()) {
      StringBuffer errorMsg = null;
      for (Problem p : parse.getProblems()) {
        if (p.getSeverity().equals(Problem.SEVERITY_ERROR) || p.getSeverity().equals(Problem.SEVERITY_FATALERROR)) {
          if (errorMsg == null) {
            errorMsg = new StringBuffer();
          }
          errorMsg.append(Misc.LINE_SEPARATOR).append("  ").append(p.toString());
          if (p.getCause() != null) {
            log.log(Level.SEVERE, p.toString(), p.getCause());
          } else {
            log.severe(p.toString());
          }
        } else {
          log.warning(p.toString());
        }
      }
      if (errorMsg != null) {
        throw new InvalidEnvironmentException("errors during parsing of " + description + ": " + errorMsg);
      }
    }
  }
}
