/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.net.URL;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.deployment.Deployment;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.remote.RemoteManagementAPI;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;

/**
 * Workflow process deployment operations. Individual or grouped deployment of objects relating to the process definition:
 * XPDL file, java class data for hooks, mappers, performer assignments, ....
 * @see org.ow2.bonita.definition.TxHook
 * @see org.ow2.bonita.definition.Hook
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public interface ManagementAPI extends RemoteManagementAPI {

  /**
   * To deploy the XPDL Workflow Package giving the URL of the XPDL file.
   * @param xpdlURL the URL of the XPDL file.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if XPDL file cannot be found at the given URL</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed.</li>
   * </ul>
   * <li>if required classes have not been found into the global class repository.</li>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployXpdl(final URL xpdlURL) throws DeploymentException;
  
  /**
   * To deploy the XPDL Workflow Package giving the byte array of the serialazed XPDL file.
   * @param xpdlFile the byte array of the XPDL file.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if XPDL file cannot be deserialized from the given byte arry</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed.</li>
   * </ul>
   * <li>if required classes have not been found into the global class repository.</li>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployXpdl(final byte[] xpdlFile) throws DeploymentException;
  
  /**
   * To deploy the XPDL Workflow Package and its depending java classes.
   * giving the URL of the XPDL file and the table of classes.<br>
   * Classes are deployed at package level (not global level repository).
   * @param xpdlURL the URL of the XPDL file.
   * @param classes the table of java classes.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if XPDL file cannot be found at the given URL</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed.</li>
   * </ul>
   * <li>if required classes have not been found firstly within the provided classes (package dependence)
   * or secondly into the global class repository.</li>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deploy(final URL xpdlURL, final Class< ? >[] classes) throws DeploymentException;

  /**
   * To deploy the XPDL Workflow giving the URL of the zip archive
   * containing the XPDL Workflow Package and optionally its depending java classes.<br>
   * Classes are deployed at package level (not global level repository).
   * @param zipURL the URL of the zip file.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if the archive does not contain any XPDL file</li>
   * <li>if the archive contains more than one XPDL file</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed.</li>
   * <li>if required classes have not been found firstly within the archive (package dependence)
   * or secondly into the global class repository.</li>
   * </ul>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployZip(final URL zipURL) throws DeploymentException;

  /**
   * To deploy the XPDL Workflow giving the byte table of the zip archive
   * containing the XPDL Workflow Package and optionnally its depending java classes.<br>
   * Classes are deployed at package level (not global level repository).
   * @param zipFile the bytes table of the zip file.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if the archive does not contain any XPDL file</li>
   * <li>if the archive contains more than one XPDL file</li>
   * <li>if a process has already been deployed with the same version. New version is required.</li>
   * <li>if the version of the package containing processes with new versions has not been changed too.</li>
   * <li>if required classes have not been found firstly within the archive (package dependence)
   * or secondly into the global class repository.</li>
   * </ul>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployZip(final byte[] zipFile) throws DeploymentException;

  /**
   * To deploy the XPDL Workflow giving the deployment object containing
   * the XPDL Workflow Package and optionnally its depending java classes.<br>
   * Classes are deployed at package level (not global level repository).
   * @param deployment the deployment object that should contain the XPDL file and the depending classes.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws DeploymentException
   * <ul>
   * <li>if the deployment does not contain any XPDL file</li>
   * <li>if the deployment contains more than one XPDL file</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed too.</li>
   * <li>if required classes have not been found firstly within the deployment object (package dependence)
   * or secondly into the global class repository.</li>
   * </ul>
   * @throws BonitaInternalException if an exception occurs.
   */
  Map<String, ProcessDefinition> deploy(final Deployment deployment) throws DeploymentException;
  /**
   * To deploy an XPDL workflow package giving the URL of archive file (.bar file)
   * containing the XPDL definition file and optionally the classes to be deployed.
   * @param barFileURL the URL of the bar file.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws IllegalArgumentException if the barFileURL is null parameter.
   * @throws DeploymentException
   * <ul>
   *
   * <li>if the barFile does not contain any XPDL file</li>
   * <li>if the archive contains more than one XPDL file</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed too.</li>
   * <li>if the required classes have not been found firstly within the archive (package dependence)
   * or secondly into the global class repository.</li>
   * </ul>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployBar(final URL barFileURL) throws DeploymentException;
  /**
   * To deploy a package giving the URL of the bar file.
   * @param barFile the file of the barfile.
   * @return a map of deployed process definitions <br>
   * (keys are the id of the processes as specified by the value of the Id attribute
   * within the WorkflowProcess elements into the XPDL definition file
   * and values are the {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition ProcessDefinition} objects).
   * @throws IllegalArgumentException if the barFile is null parameter.
   * @throws DeploymentException
   * <ul>
   * <li>if the barFile does not contain any XPDL file</li>
   * <li>if the archive contains more than one XPDL file</li>
   * <li>if a process has already been deployed with the same version. New version is required</li>
   * <li>if the version of the package containing processes with new versions has not been changed too.</li>
   * <li>if the required classes have not been found firstly within the archive (package dependence)
   * or secondly into the global class repository.</li>
   * </ul>
   * @throws BonitaInternalException if an other exception occurs.
   */
  Map<String, ProcessDefinition> deployBar(final byte[] barFile) throws DeploymentException;
  /**
   * To deploy a class giving its bytes table.
   * @param clazz the bytes table of the class.
   * @throws DeploymentException if there is already a deployed class with this name.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void deployClass(byte[] clazz) throws DeploymentException;
  /**
   * To deploy several classes in global class repository giving a set of data classes.
   * @param classes a set of classes. Each class is represented by a bytes table.
   * @throws IllegalArgumentException if classes is null parameter.
   * @throws DeploymentException if there is already a deployed class with the name
   * @throws BonitaInternalException if an other exception occurs.
   */
  void deployClasses(Set<byte[]> classes) throws DeploymentException;

  /**
   * To deploy several classes in global class repository giving an archive containing the classes.
   * @param classesArchive the archive containing the classes represented by a bytes table.
   * @throws IllegalArgumentException if the classes is null parameter.
   * @throws DeploymentException if there is already a deployed class with this name or there's an IOException occurs.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void deployClassesInJar(byte[] classesArchive) throws DeploymentException;
  /**
   * To undeploy a Workflow package giving the package UUID. This operation undeploys also
   * the classes that have been deployed within the initial package deployment.
   * @param packageUUID the package definition UUID.
   * This UUID can be retrieved from any {@link org.ow2.bonita.facade.def.majorElement.DefinitionRecord}
   * (e.g. {@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition})
   * by calling {@link org.ow2.bonita.facade.def.majorElement.DefinitionRecord#getPackageDefinitionUUID()}
   * @throws IllegalStateException if the given parameter is null.
   * @throws DeploymentException
   * <ul>
   * <li>if the package has not been found in the journal (ie. the package is not deployed)</li>
   * <li>if there's still running instances for process(es) deployed within the package.</li>.
   * </ul>
   * @throws BonitaInternalException if an exception occurs.
   */
  void undeploy(final PackageDefinitionUUID packageUUID) throws DeploymentException;
  /**
   * To remove a class giving the class name.
   * @param className the name of the class.
   * @throws DeploymentException if there's no class defined in global class repository with this name
   * or a deployed process is still using this class.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void removeClass(String className) throws DeploymentException;
  /**
   * To remove classes giving there class names.
   * @param classNames the table of class names.
   * @throws DeploymentException if there's no class defined in global class repository with this name
   * or a deployed process is still using given classes (deployed into the global class repository).
   * @throws BonitaInternalException if an other exception occurs.
   */
  void removeClasses(String[] classNames) throws DeploymentException;

  /**
   * Remove the class with the given className name. Deploy the new given one.
   * @param className the class name to be replaced.
   * @param newClazz the bytes table of the new class.
   * @throws DeploymentException if the class to be replaced is not found
   * into the global class repository.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void replaceClass(String className, byte[] newClazz) throws DeploymentException;

  /**
   * Clear the history data (archived data).
   * @throws BonitaInternalException if an other exception occurs.
   */
  void clearHistory();
  
  /**
   * Deletes from journal and history :
   * - the given package
   * - all its processes
   * - all instances of these processes.
   * Warning, UndeletablePackageException is thrown if at least one process has running instance
   * 
   * @param pacDefinitionUUID UUID of the package to delete
   * @throws PackageNotFoundException if the package can't be found
   * @throws UndeletablePackageException if at least one process of the given package has running instance
   * @throws BonitaInternalException if an other exception occurs.
   */
  void deletePackage(PackageDefinitionUUID pacDefinitionUUID) throws PackageNotFoundException, UndeletablePackageException;
}
