/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade;

import java.util.Map;

import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.AmbiguousActivityException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.remote.RemoteRuntimeAPI;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
/**
 * To manage process, process instance and task life cycle operations as well as to set/add/update
 * variables within activity or instance.
 *
 * Default states for process, processes instances, tasks (aka manual activities) are:
 * <ul>
 * <li>{@link org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState States for process}: UNDEPLOYED, DEPLOYED</li>
 * <li>{@link org.ow2.bonita.facade.runtime.InstanceState States for process instance}: INITIAL, STARTED, FINISHED</li>
 * <li>{@link org.ow2.bonita.facade.runtime.TaskState States for task}: INITIAL, READY, EXECUTING, SUSPENDED, FINISHED</li>
 * </ul>
 * @see org.ow2.bonita.definition.TxHook
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public interface RuntimeAPI extends RemoteRuntimeAPI {

  /**
   * Creates an instance of the specified process and start the execution.
   * returned instance has STARTED state.
   * If the first activity has StartMode=manual then a task has been created.
   * If the first activity has StartMode=automatic then the automatic behavior
   * of the activity has been started.
   * @param processUUID the process UUID.
   * @return the UUID of the created instance.
   * @throws ProcessNotFoundException if the process has not been found.
   * @throws BonitaInternalException if an exception occurs.
   */
  ProcessInstanceUUID instantiateProcess(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException;

  /**
   * Creates an instance of the specified process with the added variable map
   * and start the execution.
   * returned instance has STARTED state.
   * If the first activity has StartMode=manual then a task has been created.
   * If the first activity has StartMode=automatic then the automatic behavior
   * of the activity has been started.
   * @param processUUID the process UUID.
   * @param variables variables added to the variables already set within the process definition.
   * @return the UUID of the created instance.
   * @throws ProcessNotFoundException if the process has not been found.
   * @throws BonitaInternalException if an exception occurs.
   */
  ProcessInstanceUUID instantiateProcess(ProcessDefinitionUUID processUUID, Map<String, Object> variables)
    throws ProcessNotFoundException;

  /**
   * Starts the task. If successful, this operation changes task state from READY to EXECUTING.<br>
   * If the boolean assignTask is true the task is also assigned to the logged user
   * otherwise the assignment of the task is not affected by this operation.
   * @param taskUUID the task UUID.
   * @param assignTask TODO
   * @param assignTask true to assign the task to the logged user; false to don't assign the task.
   * @throws TaskNotFoundException if the task has not been found.
   * @throws IllegalTaskStateException if the state of the task has not READY state.
   * @throws BonitaInternalException if an exception occurs.
   */
  void startTask(TaskUUID taskUUID, boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException;

  /**
   * Finishes the task. If successful, this operation changes task state from EXECUTING to FINISHED.<br>
   * If the boolean assignTask is true the task is also assigned to the logged user
   * otherwise the assignment of the task is not affected by this operation.
   * @param taskUUID the task UUID.
   * @throws TaskNotFoundException if the task has not been found.
   * @throws IllegalTaskStateException if the state of the task has not EXECUTING state.
   * @throws BonitaInternalException if an exception occurs.
   */
  void finishTask(TaskUUID taskUUID, boolean taskAssign) throws TaskNotFoundException, IllegalTaskStateException;

  /**
   * Suspends the task if the task has EXECUTING state.<br>
   * If successful, this operation changes task state from EXECUTING to SUSPENDED.<br>
   * If the boolean assignTask is true the task is also assigned to the logged user
   * otherwise the assignment of the task is not affected by this operation.
   * @param taskUUID the task UUID.
   * @throws TaskNotFoundException if the task has not been found.
   * @throws IllegalTaskStateException if the state of the task has not either READY or EXECUTING state.
   * @throws BonitaInternalException if an exception occurs.
   */
  void suspendTask(TaskUUID taskUUID, boolean assignTask) throws TaskNotFoundException,
    IllegalTaskStateException;

  /**
   * Resumes the task if the task has SUSPENDED state.
   * If successful, this operation changes task state from SUSPENDED to EXECUTING.<br>
   * If the boolean assignTask is true the task is also assigned to the logged user
   * otherwise the assignment of the task is not affected by this operation.
   * @param taskUUID the task UUID.
   * @throws TaskNotFoundException if the task has not been found.
   * @throws IllegalTaskStateException if the state of the task has not SUSPENDED state.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void resumeTask(TaskUUID taskUUID, boolean taskAssign) throws TaskNotFoundException,
    IllegalTaskStateException;

  /**
   * Launches the execution of both roleMapper and performerAssignment for the given task.<br>
   * If a roleMapper has been defined within the participant referenced by the performer of the task,
   * it is executed.<br>
   * If a performerAssignment has been defined within the activity of the task it is also
   * executed.
   * @param taskUUID the task UUID.
   * @throws TaskNotFoundException if the task has not been found.
   * @throws BonitaInternalException if an other exception occurs.
   */
  void assignTask(TaskUUID taskUUID) throws TaskNotFoundException;

  /**
   * Forces to assign the given task to the given actor id.
   * @param taskUUID the task UUID.
   * @param actorId the actor id.
   * @throws TaskNotFoundException if the task has not been found.
   */
  void assignTask(TaskUUID taskUUID, String actorId) throws TaskNotFoundException;

  /**
   * Forces to replace the candidates set of the given task by the given candidates set.
   * @param taskUUID the task UUID.
   * @param candidates the set of candidate actors.
   * @throws TaskNotFoundException if the task has not been found.
   */
  void assignTask(TaskUUID taskUUID, java.util.Set<java.lang.String> candidates) throws TaskNotFoundException;

  /**
   * Searches for variable with id variableId within the given process instance
   * with ProcessInstanceUUID instanceUUID.<br>
   * If the variable is found, the given value is set.<br>
   * <ul>
   * <li>If the XPDL dataField is String type then use instance of
   * java.lang.String class for variableValue.</li>
   * <li>If the XPDL dataField is Enumeration type then use instance of
   * {@link org.ow2.bonita.runtime.var.Enumeration Enumeration} class for variableValue.</li>
   * </ul>
   * The field selectedValue of Enumeration type can contain only one value
   * from the possibleValues list field.
   * @param instanceUUID the instance UUID.
   * @param variableId the variable id.
   * @param variableValue the variable value.
   * @throws InstanceNotFoundException if the instance has not been found.
   * @throws VariableNotFoundException if the variable has not been found.
   * @throws BonitaInternalException if an exception occurs.
   */
  void setProcessInstanceVariable(ProcessInstanceUUID instanceUUID, String variableId, Object variableValue)
    throws InstanceNotFoundException, VariableNotFoundException;

  /**
   * <p>
   * Searches for variable with id variableId within the given activity instance
   * with the given instanceUUID, activityId.<br>
   * If the variable is found within the activity, the given value is set.<br>
   * If the variable is not found within the activity the search is performed
   * within the process instance.<br>
   * If the variable is found within the process instance, the given value is set.
   * </p>
   * <ul>
   * <li>If the XPDL dataField is String type then use instance of
   * java.lang.String class for variableValue.</li>
   * <li>If the XPDL dataField is Enumeration type then use instance of
   * {@link org.ow2.bonita.runtime.var.Enumeration Enumeration} class for variableValue.</li>
   * </ul>
   * The field selectedValue of Enumeration type can contain only one value
   * from the possibleValues list field.
   * @param instanceUUID the ProcessInstanceUUID.
   * @param variableId the variable id.
   * @param variableValue the variable value.
   * @throws InstanceNotFoundException if the instance has not been found.
   * @throws VariableNotFoundException if the variable has not been found.
   * @throws AmbiguousActivityException
   * @throws BonitaInternalException if an exception occurs.
   */
  void setVariable(ProcessInstanceUUID instanceUUID, String activityId, String variableId, Object variableValue)
    throws InstanceNotFoundException, ActivityNotFoundException, VariableNotFoundException, AmbiguousActivityException;

  /**
   * Searches for variable with id variableId into instance
   * with UUID processUUID and within the activity with id activityId.<br>
   * If the activity variable is found, the given value is set.<br>
   * <ul>
   * <li>If the XPDL dataField has String type then use instance of
   * java.lang.String class for variableValue.</li>
   * <li>If the XPDL dataField has EnumerationType type then use instance of
   * {@link org.ow2.bonita.runtime.var.Enumeration Enumeration} class for variableValue.</li>
   * </ul>
   * The field selectedValue can contain only one value from the possibleValues list field.
   * @param instanceUUID the instance UUID.
   * @param activityId the activity id.
   * @param variableId the variable id.
   * @param variableValue the variable value.
   * @throws InstanceNotFoundException if the instance has not been found.
   * @throws ActivityNotFoundException if the activity has not been found.
   * @throws VariableNotFoundException if the variable has not been found.
   * @throws AmbiguousActivityException
   * @throws BonitaInternalException if an exception occurs.
   */
  void setActivityInstanceVariable(ProcessInstanceUUID instanceUUID, String activityId, String variableId, Object variableValue)
    throws InstanceNotFoundException, ActivityNotFoundException, VariableNotFoundException, AmbiguousActivityException;

  /**
   * Searches for variable with the given activity UUID and variable Id.<br>
   * If the activity variable is found, the given value is set.<br>
   * <ul>
   * <li>If the XPDL dataField has String type then use instance of
   * java.lang.String class for variableValue.</li>
   * <li>If the XPDL dataField has EnumerationType type then use instance of
   * {@link org.ow2.bonita.runtime.var.Enumeration Enumeration} class for variableValue.</li>
   * </ul>
   * The field selectedValue can contain only one value from the possibleValues list field.
   * @param activityUUID the activity UUID.
   * @param variableId the variable id.
   * @param variableValue the variable value.
   * @throws ActivityNotFoundException if the activity has not been found.
   * @throws VariableNotFoundException if the variable has not been found.
   * @throws BonitaInternalException if an exception occurs.
   */
  void setActivityInstanceVariable(ActivityInstanceUUID activityUUID,
      String variableId, Object variableValue) throws ActivityNotFoundException, VariableNotFoundException;

  /**
   * Deletes all runtime objects for the process instance with the given instance UUID
   * and delete also recorded data from the journal.
   * If this instance was not found in the journal, then the archived instance is deleted from history.
   * @param instanceUUID the instance UUID.
   * @throws InstanceNotFoundException if if the instance has not been found.
   * @throws BonitaInternalException if an exception occurs.
   */
  void deleteProcessInstance(ProcessInstanceUUID instanceUUID) throws InstanceNotFoundException;

  /**
   * Deletes all runtime objects for all instances created with the given process UUID
   * and delete also all there recorded data from the journal.
   * If instances some instances of this process were not found in the journal,
   * then the archived instances are deleted from history.
   * @param processUUID the process UUID.
   * @throws ProcessNotFoundException if the process with the given UUID does not exists.
   * @throws BonitaInternalException if an exception occurs.
   */
  void deleteAllProcessInstances(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException;

}


