/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PackageDependency;
import org.ow2.bonita.deployment.Deployer;
import org.ow2.bonita.deployment.Deployment;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.PackageNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletablePackageException;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.parsing.XpdlParserException;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Tool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class ManagementAPIImpl implements ManagementAPI {

  private static Logger log = Logger.getLogger(ManagementAPIImpl.class.getName());

  public Map<String, ProcessDefinition> deploy(final Deployment deployment) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(deployment);
    try {
      return Deployer.deploy(deployment, EnvTool.getUserId());
    } catch (DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    } catch (XpdlParserException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public Map<String, ProcessDefinition> deploy(URL xpdlURL, Class< ? >[] classes) throws DeploymentException {
    return deploy(new Deployment(xpdlURL, Tool.getClasses(classes)));
  }

  public Map<String, ProcessDefinition> deployXpdl(byte[] xpdlFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(xpdlFile);
    return deploy(new Deployment(xpdlFile, null));
  }

  public Map<String, ProcessDefinition> deployXpdl(URL xpdlURL) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(xpdlURL);
    return deploy(new Deployment(xpdlURL, null));
  }

  public Map<String, ProcessDefinition> deployZip(byte[] zipFile) throws DeploymentException {
    return deployArchive(zipFile);
  }

  public Map<String, ProcessDefinition> deployZip(URL zipURL) throws DeploymentException {
    return deployArchive(zipURL);
  }

  public Map<String, ProcessDefinition> deployBar(URL jarFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(jarFile);
    return deployArchive(jarFile);
  }

  public Map<String, ProcessDefinition> deployBar(byte[] barFile) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(barFile);
    return deployArchive(barFile);
  }

  public void deployClass(byte[] clazz) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(clazz);
    String className;
    try {
      className = Tool.visitClass(clazz).getClassName();
    } catch (ArrayIndexOutOfBoundsException e) {
      throw new DeploymentException("The byte table is not correct parameter.");
    }
    Repository repository = EnvTool.getRepository();
    GlobalClassData globalClassData = repository.getGlobalClassData(className);
    if (globalClassData != null) {
      //this class was already deployed
      throw new DeploymentException("There is already a deployed class with this name: ", className);
    }
    repository.storeGlobalClassData(new GlobalClassData(className, clazz));
  }

  public void deployClassesInJar(byte[] classesArchive) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classesArchive);
    File tmpDir = explodeArchive(classesArchive, "globalClasses");
    Set<byte[]> classes = getClasses(tmpDir);
    deployClasses(classes);
  }

  public void deployClasses(Set<byte[]> classes) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classes);
    for (byte[] data : classes) {
      deployClass(data);
    }
  }

  public void undeploy(final PackageDefinitionUUID packageUUID) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(packageUUID);
    try {
      Deployer.undeployPackage(packageUUID, EnvTool.getUserId());
    } catch (DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void removeClass(String className) throws DeploymentException {
    Repository repository = EnvTool.getRepository();
    if (repository.getGlobalClassData(className) == null) {
      throw new DeploymentException("There is no class defined in global class repository  with name: ", className);

    }
    // check that no process is still using this class
    Set<PackageDependency> packageDependencies = repository.getPackageDependencies(className);
    if (packageDependencies != null && !packageDependencies.isEmpty()) {
      for (PackageDependency packageDependency : packageDependencies) {
        PackageClassData data = repository.getPackageClassData(packageDependency.getPackageDefinitionUUID());
        if (data == null || !data.getClasses().containsKey(className)) {
          throw new DeploymentException("Package is still using global class. "
              + "Cannot undeploy class.", className , packageDependency.getPackageDefinitionUUID());
        }
      }
    }
    repository.removeGlobalClassData(className);
    ClassDataLoader.removeGlobalClassLoader(className);
  }

  public void removeClasses(String[] classNames) throws DeploymentException {
    if (classNames != null) {
      for (String className : classNames) {
        removeClass(className);
      }
    }
  }

  public void replaceClass(String className, byte[] newClazz) throws DeploymentException {
    removeClass(className);
    deployClass(newClazz);
  }

  public void clearHistory() {
    final Archiver archiver = EnvTool.getArchiver();
    archiver.clear();
  }

  private Map<String, ProcessDefinition> deployArchive(URL archiveURL) throws DeploymentException {
    InputStream stream = null;
    try {
      stream = archiveURL.openStream();
      return deployArchive(Misc.getAllContentFrom(stream));
    } catch (IOException e) {
      throw new DeploymentException("Problem while deploying bar: ", e);
    } finally {
      Misc.close(stream);
    }
  }

  private Map<String, ProcessDefinition> deployArchive(byte[] barFile) throws DeploymentException {
    File tmpDir;
    tmpDir = explodeArchive(barFile, "tmpBonitaBar");
    List<File> xpdlFiles = getMatchingFiles(tmpDir, ".xpdl");
    if (xpdlFiles.isEmpty()) {
      throw new DeploymentException("The given bar archive does not contain any '.xpdl' file");
    }
    if (xpdlFiles.size() > 1) {
      throw new DeploymentException("The given bar archive contains many '.xpdl' files");
    }
    URL xpdlURL;
    try {
      xpdlURL = xpdlFiles.get(0).toURL();
      if (log.isLoggable(Level.INFO)) {
        log.info("Xpdl file currently deployed: " + xpdlURL);
      }
    } catch (MalformedURLException e) {
      throw new DeploymentException("Error when deploying the bar file", e);
    }
    Map<String, ProcessDefinition> processes = deploy(new Deployment(xpdlURL, getClasses(tmpDir)));
    Misc.deleteDir(tmpDir);
    return processes;
  }

  private String getClassName(File classFile, File baseDir) {
    String basePath = baseDir.getAbsolutePath();
    String classPath = classFile.getAbsolutePath();
    String className = classPath.substring(basePath.length() + 1);
    String suffix = ".class";
    className = className.substring(0, className.length() - suffix.length());
    return className.replace(File.separatorChar, '.');
  }

  private List<File> getMatchingFiles(File baseDir, String extension) {
    List<File> files = new ArrayList<File>();
    getMatchingFiles(baseDir, files, extension);
    return files;
  }

  private void getMatchingFiles(File dir, List<File> matchingFiles, String extension) {
    for (File f : dir.listFiles()) {
      if (f.getName().endsWith(extension)) {
        matchingFiles.add(f);
      } else if (f.isDirectory()) {
        getMatchingFiles(f, matchingFiles, extension);
      }
    }
  }

  private File explodeArchive(byte[] archive, String destDirName) throws DeploymentException {
    File tmpDir = null;
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      tmpDir = File.createTempFile(destDirName, null, null);
      Misc.badStateIfFalse(tmpDir.delete(), "Can't delete: " + tmpDir);
      if (!tmpDir.mkdirs()) {
        throw new DeploymentException("Cannot create the temporary directory:' " + tmpDir);
      }

      // Unzip file
      InputStream in = new ByteArrayInputStream(archive);
      ZipInputStream zis = null;
      zis = new ZipInputStream(in);

      ZipEntry zipEntry = null;
      while ((zipEntry = zis.getNextEntry()) != null) {
        // write the file to the disk
        // Create file
        File outputFile = new File(tmpDir.getAbsolutePath() + File.separator + zipEntry.getName());
        if (zipEntry.isDirectory()) {
          outputFile.mkdirs();
        } else {
          int i = 0;
          int position = 0;
          while (i < zipEntry.getName().length()) {
            if (zipEntry.getName().charAt(i) == File.separatorChar) {
              position = i + 1;
            }
            i++;
          }
          new File(tmpDir.getAbsolutePath() + File.separator + zipEntry.getName().substring(0, position)).mkdirs();
          final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(outputFile));
          int c;
          byte[] buffer = new byte[512];
          while ((c = zis.read(buffer)) != -1) {
            bos.write(buffer, 0, c);
          }
          //bos.flush();
          Misc.close(bos);
        }
      }
      zis.close();
      in.close();
    } catch (IOException e) {
      throw new DeploymentException("Problem while exploding jar file at deployment", e);
    }
    return tmpDir;
  }

  private Set<byte[]> getClasses(File tmpDir) throws DeploymentException {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Looking for datas in directory: " + tmpDir);
    }
    List<File> classFiles = getMatchingFiles(tmpDir, ".class");
    Set<byte[]> classes = new HashSet<byte[]>();
    try {
      for (File classFile : classFiles) {
        String className = getClassName(classFile, tmpDir);
        URL classURL = classFile.toURL();
        if (log.isLoggable(Level.FINE)) {
          log.fine("Found class: " + className + " with URL: " + classURL);
        }
        classes.add(Misc.getAllContentFrom(classURL));
      }
    } catch (IOException e) {
      throw new DeploymentException("Error when getting classDatas at deployment ", e);
    }
    return classes;
  }

  public void deletePackage(PackageDefinitionUUID packageDefinitionUUID) throws
  PackageNotFoundException, UndeletablePackageException {
    FacadeUtil.checkArgsNotNull(packageDefinitionUUID);

    Querier journal = EnvTool.getJournalQueriers();
    
    PackageFullDefinition packageDefinition = journal.getPackage(packageDefinitionUUID);
    boolean inJournal = packageDefinition != null;
    if (!inJournal) {
      Querier history = EnvTool.getHistoryQueriers();
      packageDefinition = history.getPackage(packageDefinitionUUID);
    }
    
    if (packageDefinition == null) {
      throw new PackageNotFoundException(packageDefinitionUUID);
    }
    Repository repository = EnvTool.getRepository();
    Set<ProcessFullDefinition> processes = packageDefinition.getProcesses();

    //check that none of the processes have running instances
    for (ProcessFullDefinition process : processes) {
      ProcessDefinitionUUID processUUID = process.getUUID();
      Set<XpdlInstance> xpdlInstances = repository.getXpdlInstances(processUUID);
      if (!xpdlInstances.isEmpty()) {
        ProcessInstanceUUID instanceUUID = xpdlInstances.iterator().next().getUUID();
        throw new UndeletablePackageException(packageDefinitionUUID, processUUID, instanceUUID);
      }

    }
    RuntimeAPI runtimeAPI = new RuntimeAPIImpl();
    try {
      for (ProcessFullDefinition process : processes) {
        ProcessDefinitionUUID processUUID = process.getUUID();
        runtimeAPI.deleteAllProcessInstances(processUUID);
      }
    } catch (ProcessNotFoundException e) {
      throw new BonitaRuntimeException(e);
    }
    if (inJournal) {
      Deployer.removePackageDependencies(packageDefinition);
    }
    if (packageDefinition.getUndeployedBy() != null) {
      Archiver archiver = EnvTool.getArchiver();
      archiver.remove(packageDefinition);
    } else {
      Recorder recorder = EnvTool.getRecorder();
      recorder.remove(packageDefinition);
    }
  }
}
