/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;


import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.job.Timer;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.dataType.BasicTypeDefinition;
import org.ow2.bonita.facade.def.dataType.DataTypeDefinition;
import org.ow2.bonita.facade.def.element.FormalParameterDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.exception.ActivityNotFoundException;
import org.ow2.bonita.facade.exception.AmbiguousActivityException;
import org.ow2.bonita.facade.exception.BonitaInternalException;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.InstanceNotFoundException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.exception.TypeMismatchException;
import org.ow2.bonita.facade.exception.VariableNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.runtime.var.Enumeration;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.EnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */

public class RuntimeAPIImpl implements RuntimeAPI {
  private static final Logger LOG = Logger.getLogger(ManagementAPIImpl.class.getName());

  /**
   * Create an instance of the specified process and return the processUUID
   */
  public ProcessInstanceUUID instantiateProcess(final ProcessDefinitionUUID processId) throws ProcessNotFoundException {
    return instantiateProcess(processId, null);
  }

  public ProcessInstanceUUID instantiateProcess(final ProcessDefinitionUUID processId,
      Map<String, Object> variables) throws ProcessNotFoundException {
    FacadeUtil.checkArgsNotNull(processId);
    final XpdlProcess process = FacadeUtil.getProcessDefinition(processId);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.finest("Starting a new instance of process : " + processId);
    }
    final XpdlExecution rootExecution = (XpdlExecution) process.beginProcessInstance();
    rootExecution.setVariables(variables);
    rootExecution.getExecution(process.getInitial().getName()).signal();

    if (LOG.isLoggable(Level.FINE)) {
      LOG.finest("Started: " + rootExecution.getXpdlInstance());
    }
    return rootExecution.getXpdlInstance().getUUID();
  }

  public void deleteProcessInstance(ProcessInstanceUUID instanceUUID) throws InstanceNotFoundException {
    Repository repository = EnvTool.getRepository();
    Querier journal = EnvTool.getJournalQueriers();
    ProcessFullInstance processInst = journal.getProcessInstance(instanceUUID);
    boolean inJournal = processInst != null;
    if (inJournal) {
      List<Timer> timers = repository.getInstanceTimers(instanceUUID);
      if (timers != null) {
        for (Timer timer : timers) {
          repository.removeTimer(timer);
        }
      }
      List<TaskRunTime> tasks = repository.getInstanceTasks(instanceUUID);
      if (tasks != null) {
        for (TaskRunTime task : tasks) {
          repository.removeTask(task);
        }
      }
      repository.removeXpdlInstance(instanceUUID);
      Recorder recorder = EnvTool.getRecorder();
      recorder.remove(processInst);
      return;
    }
    Querier history = EnvTool.getHistoryQueriers();
    processInst = history.getProcessInstance(instanceUUID);
    boolean inHistory = processInst != null;
    if (inHistory) {
      Archiver archiver = EnvTool.getArchiver();
      archiver.remove(processInst);
      return;
    }
    throw new InstanceNotFoundException(instanceUUID);    
  }
  public void deleteAllProcessInstances(ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    Querier querier = EnvTool.getAllQueriers();
    ProcessFullDefinition process = querier.getProcess(processUUID);
    if (process == null) {
      throw new ProcessNotFoundException(processUUID);
    }
    
    Set<ProcessFullInstance> instances = querier.getProcessInstances(processUUID);
    
    for (ProcessFullInstance instance : instances) {
      try {
        deleteProcessInstance(instance.getUUID());
      } catch (InstanceNotFoundException e) {
        throw new BonitaInternalException("Invalid state", e);
      }
    }
  }

  public void startTask(final TaskUUID taskUUID, boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException  {
    FacadeUtil.checkArgsNotNull(taskUUID);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }

    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Checking compatible state of " + taskRunTime);
    }
    if (!taskRunTime.getState().equals(TaskState.READY)) {
      Set<TaskState> expectedStates = new HashSet<TaskState>();
      expectedStates.add(TaskState.READY);
      throw new IllegalTaskStateException("Can't start task with processDefinitionUUID = ", taskRunTime.getUUID(),
          expectedStates, taskRunTime.getState());
    }
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Starting: " + taskRunTime);
    }
    taskRunTime.start(EnvTool.getUserId(), assignTask);
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Started: " + taskRunTime);
    }
  }

  public void finishTask(TaskUUID taskUUID, boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }

    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Checking compatible state of " + taskRunTime);
    }
    if (!taskRunTime.getState().equals(TaskState.EXECUTING)) {
      Set<TaskState> expectedStates = new HashSet<TaskState>();
      expectedStates.add(TaskState.EXECUTING);
      throw new IllegalTaskStateException("Can't finish task with processDefinitionUUID = ",
          taskRunTime.getUUID(), expectedStates, taskRunTime.getState());
    }
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Terminating: " + taskRunTime);
    }
    taskRunTime.finish(EnvTool.getUserId(), assignTask);
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Terminated: " + taskRunTime);
    }
  }

  public void suspendTask(final TaskUUID taskUUID, boolean assignTask) throws TaskNotFoundException,
  IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }

    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Checking compatible state of " + taskRunTime);
    }
    if (!(taskRunTime.getState().equals(TaskState.READY)
        || taskRunTime.getState().equals(TaskState.EXECUTING))) {
      Set<TaskState> expectedStates = new HashSet<TaskState>();
      expectedStates.add(TaskState.READY);
      expectedStates.add(TaskState.EXECUTING);
      throw new IllegalTaskStateException("Can't suspend task with processDefinitionUUID = ",
          taskRunTime.getUUID(), expectedStates, taskRunTime.getState());
    }

    String currentUserId = EnvTool.getUserId();
    /* patch
    //If the suspended task is already assigned to a user (userId is set)
    //it can only be resumed by this user
    String taskuserId = taskRunTime.getUserId();
    if (taskuserId != null && !taskuserId.equals(currentUserId)) {
      throw new UnAuthorizedUserException("to suspend the task", taskRunTime.getUUID(), currentUserId);
    } else if (taskuserId == null) {
      Set<String> candidates = taskRunTime.getCandidates();
      if (candidates == null || !candidates.contains(currentUserId)) {
        throw new UnAuthorizedUserException("to suspend the task", taskRunTime.getUUID(), currentUserId);
      }
    }
    */
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Suspending: " + taskRunTime);
    }
    taskRunTime.suspend(currentUserId, assignTask);
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Suspended: " + taskRunTime);
    }
  }

  public void resumeTask(final TaskUUID taskUUID, boolean taskAssign) throws TaskNotFoundException,
  IllegalTaskStateException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }

    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Checking compatible state of " + taskRunTime);
    }
    if (!taskRunTime.getState().equals(TaskState.SUSPENDED)) {
      Set<TaskState> expectedStates = new HashSet<TaskState>();
      expectedStates.add(TaskState.SUSPENDED);
      throw new IllegalTaskStateException("Can't suspend task with processDefinitionUUID = ",
          taskRunTime.getUUID(), expectedStates, taskRunTime.getState());
    }
    String currentUserId = EnvTool.getUserId();
    /*
    //If the suspended task is already assigner to a user (userId is set)
    //it can only be resumed by this user
    String taskUserId = taskRunTime.getUserId();
    if (taskUserId != null && !taskUserId.equals(currentUserId)) {
      throw new UnAuthorizedUserException("to resume the task", taskRunTime.getUUID(), currentUserId);
    } else if (taskUserId == null) {
      Set<String> candidates = taskRunTime.getCandidates();
      if (candidates == null || !candidates.contains(currentUserId)) {
        throw new UnAuthorizedUserException("to resume the task", taskRunTime.getUUID(), currentUserId);
      }
    }
    */
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Resuming: " + taskRunTime);
    }
    taskRunTime.resume(currentUserId, taskAssign);
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Resumed: " + taskRunTime);
    }
  }

  public void assignTask(TaskUUID taskUUID) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }
    Performer performer = taskRunTime.getActivityDef().getPerformer();
    //execute Role mapper and performerAssign
    if (performer.getRoleMapper() != null) {
      taskRunTime.resolve(performer);
    }

    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigning: " + taskRunTime);
    }
    //record the assign
    taskRunTime.assign(taskRunTime.getCandidates(), taskRunTime.getUserId());
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigned: " + taskRunTime);
    }
  }

  public void assignTask(TaskUUID taskUUID, String userId) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID, userId);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigning: " + taskRunTime);
    }
    //record the assign
    taskRunTime.assign(taskRunTime.getCandidates(), userId);
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigned: " + taskRunTime);
    }
  }

  public void assignTask(TaskUUID taskUUID, java.util.Set<java.lang.String> candidates) throws TaskNotFoundException {
    FacadeUtil.checkArgsNotNull(taskUUID, candidates);
    Repository repository = EnvTool.getRepository();
    TaskRunTime taskRunTime = repository.getTask(taskUUID);
    if (taskRunTime == null) {
      throw new TaskNotFoundException(taskUUID);
    }
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigning: " + taskRunTime);
    }
    //record the assign
    taskRunTime.assign(candidates, taskRunTime.getUserId());
    if (LOG.isLoggable(Level.FINEST)) {
      LOG.finest("Assigned: " + taskRunTime);
    }
  }

  public void setProcessInstanceVariable(ProcessInstanceUUID instanceUUID,
      String variableId, Object variableValue) throws InstanceNotFoundException, VariableNotFoundException {
    final XpdlInstance instance = FacadeUtil.getInstance(instanceUUID);
    if (instance == null) {
      throw new InstanceNotFoundException(instanceUUID);
    }
    XpdlExecution rootExecution = instance.getRootExecution();
    Map<String, Object> globalVars = rootExecution.getScopeVariables();
    if (!globalVars.containsKey(variableId)) {
      throw new VariableNotFoundException(instanceUUID, variableId);
    }
    ProcessDefinitionUUID processDefinitionUUID = instance.getProcessDefinitionUUID();
    PackageDefinitionUUID packageDefinitionUUID = instance.getPackageDefinitionUUID();
    ProcessFullDefinition processFullDefinition = EnvTool.getAllQueriers().getProcess(processDefinitionUUID);
    Misc.badStateIfNull(processFullDefinition, "can't find a process with uuid " + processDefinitionUUID);


    DataTypeDefinition dataType = null;
    for (DataFieldDefinition df : processFullDefinition.getDataFields()) {
      if (df.getDataFieldId().equals(variableId)) {
        dataType = df.getDataType();
      }
    }
    if (dataType == null) {
      for (FormalParameterDefinition formalParameterDefinition : processFullDefinition.getFormalParameters()) {
        if (formalParameterDefinition.getId().equals(variableId)) {
          dataType = formalParameterDefinition.getDataType();
        }
      }
    }
    if (dataType == null) {
      PackageFullDefinition packageFullDefinition = EnvTool.getAllQueriers().getPackage(packageDefinitionUUID);
      Misc.badStateIfNull(packageFullDefinition, "can't find a package with uuid " + packageDefinitionUUID);
      for (DataFieldDefinition df : processFullDefinition.getDataFields()) {
        if (df.getDataFieldId().equals(variableId)) {
          dataType = df.getDataType();
        }
      }
    }
    Misc.badStateIfNull(dataType, "can't find a datafield for variable " + variableId);
    checkAssignmentCompatibility(variableId, variableValue, dataType, processDefinitionUUID, packageDefinitionUUID);

    rootExecution.setVariable(variableId, variableValue);
    EnvTool.getRecorder().recordInstanceVariableUpdated(variableId, variableValue, instance.getUUID(), EnvTool.getUserId());
  }

  public void setActivityInstanceVariable(ProcessInstanceUUID instanceUUID, String activityId, String variableId,
      Object variableValue) throws InstanceNotFoundException, ActivityNotFoundException,
      VariableNotFoundException, AmbiguousActivityException {
    ActivityInstance<ActivityBody> activity = new QueryRuntimeAPIImpl().getActivityInstance(instanceUUID, activityId);
    setActivityInstanceVariable(activity.getUUID(), variableId, variableValue);
  }

  public void setActivityInstanceVariable(ActivityInstanceUUID activityUUID,
      String variableId, Object variableValue) throws ActivityNotFoundException, VariableNotFoundException {
    ActivityFullInstance<ActivityBody> activity = EnvTool.getAllQueriers().getActivityInstance(activityUUID);
    if (activity == null) {
      throw new ActivityNotFoundException(activityUUID);
    }
    ProcessInstanceUUID instanceUUID = activity.getProcessInstanceUUID();
    String activityId = activity.getActivityId();

    final Repository repository = EnvTool.getRepository();
    XpdlExecution execution = repository.getExecutionPointingOnNode(instanceUUID, activityId);
    if (execution == null) {
      throw new ActivityNotFoundException(activityUUID);
    }
    if (execution.hasExecution(activityId)) {
      execution = (XpdlExecution) execution.getExecution(activityId);
    }
    final Recorder recorder = EnvTool.getRecorder();
    if (!execution.getScopeVariables().containsKey(variableId)) {
      throw new VariableNotFoundException(instanceUUID, activityId, variableId);
    }

    ProcessDefinitionUUID processDefinitionUUID = activity.getProcessDefinitionUUID();
    PackageDefinitionUUID packageDefinitionUUID = activity.getPackageDefinitionUUID();
    ActivityDefinition activityDefinition;
    try {
      activityDefinition = new QueryDefinitionAPIImpl().getProcessActivity(processDefinitionUUID, activityId);
    } catch (ProcessNotFoundException e) {
      throw new IllegalStateException(e);
    }

    DataTypeDefinition dataType = null;
    for (DataFieldDefinition df : activityDefinition.getDataFields()) {
      if (df.getDataFieldId().equals(variableId)) {
        dataType = df.getDataType();
      }
    }
    Misc.badStateIfNull(dataType, "can't find a datafield for variable " + variableId);
    checkAssignmentCompatibility(variableId, variableValue, dataType, processDefinitionUUID, packageDefinitionUUID);
    execution.setVariable(variableId, variableValue);

    // local variable updated -> update only current activity
    recorder.recordActivityVariableUpdated(variableId, variableValue, execution.getXpdlInstance().getUUID(),
        execution.getNode().getName(), execution.getIterationId(), execution.getActivityInstanceId(), EnvTool.getUserId());
  }

  public void setVariable(ProcessInstanceUUID instanceUUID, String activityId,
      String variableId, Object variableValue) throws InstanceNotFoundException, ActivityNotFoundException,
      VariableNotFoundException, AmbiguousActivityException {
    try {
      setActivityInstanceVariable(instanceUUID, activityId, variableId, variableValue);
    } catch (VariableNotFoundException e) {
      setProcessInstanceVariable(instanceUUID, variableId, variableValue);
    }
  }

  // Checks if the value can be assigned to the variable, throws an exception otherwise.
  private void checkAssignmentCompatibility(String variableId, Object variableValue, DataTypeDefinition dataType,
      ProcessDefinitionUUID processUUID, PackageDefinitionUUID packageUUID) {
    if (!isAssignmentCompatible(variableValue, dataType, processUUID, packageUUID)) {
      throw new BonitaWrapperException(
          new TypeMismatchException("Invalid value: '" + variableValue + "' for variable '" + variableId + "'")
      );
    }
  }

  // Returns true is the object variable value can be assigned to a variable of type dataType.
  private boolean isAssignmentCompatible(Object variableValue, DataTypeDefinition dataType,
      ProcessDefinitionUUID processUUID, PackageDefinitionUUID packageUUID) {
    switch (dataType.getType()) {
      case BasicType:
        BasicTypeDefinition basicTypeDef = (BasicTypeDefinition) dataType.getValue().copy();
        switch (basicTypeDef.getType()) {
          case BOOLEAN:
            return variableValue instanceof Boolean;
          case DATETIME:
            return variableValue instanceof Date;
          case PERFORMER :
            if (!(variableValue instanceof String)) {
              return false;
            }
            ProcessFullDefinition processDefinition = EnvTool.getAllQueriers().getProcess(processUUID);
            PackageFullDefinition packageDefinition = EnvTool.getAllQueriers().getPackage(packageUUID);
            for (ParticipantDefinition p : processDefinition.getParticipants()) {
              if (p.getParticipantId().equals(variableValue)) {
                return true;
              }
            }
            for (ParticipantDefinition p : packageDefinition.getParticipants()) {
              if (p.getParticipantId().equals(variableValue)) {
                return true;
              }
            }
            return false;
          case FLOAT :
            // xpdl spec says a float can b a float or a double, in bonita it is a Double
            return variableValue instanceof Double;
          case INTEGER :
            return variableValue instanceof Integer;
          case STRING :
            return variableValue instanceof String;
          default:
            return false;
        }
      case EnumerationType:
        return variableValue instanceof Enumeration;
      default:
        throw new UnsupportedOperationException();
    }
  }

}


