/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.runtime.impl;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;

import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.AssignUpdate;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.runtime.Update;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.Tool;
import org.ow2.novabpm.util.Misc;
import org.ow2.novabpm.util.Misc.NullCheckResult;

/**
 * @author Pierre Vigneras
 * @date Mar 13, 2008
 */
public class TaskInstanceImpl extends ActivityBodyImpl implements TaskInstance {
  private static final long serialVersionUID = 4221520779461770199L;
  protected TaskState initialState;
  protected String initialUserId;
  protected Set<String> initialCandidates;

  protected String startedBy;
  protected String endedBy;
  protected Date dueDate;
  protected Date createdDate;
  protected List<Update> updates;

  protected TaskInstanceImpl() {
    super();
  }

  protected TaskInstanceImpl(TaskUUID taskUUID, String initialUserId,
      Set<String> initialCandidates, TaskState initialState, Date createdDate, Date dueDate, List<Update> updates) {
    super(taskUUID);
    NullCheckResult checkResult = Misc.findNull(taskUUID, initialState, createdDate);
    if (checkResult.hasNull()) {
      List<String> message = Misc.getStringFrom(checkResult, "taskUUID", "initialState", "createdDate");
      throw new IllegalArgumentException("These arguments are null: " + message);
    }
    this.initialUserId = initialUserId;
    this.initialCandidates = initialCandidates;
    this.initialState = initialState;
    this.createdDate = createdDate;
    this.dueDate = dueDate;
    this.updates = updates;
  }

  /**
   * @param processDefinitionUUID
   * @param deploymentId
   */
  public TaskInstanceImpl(final TaskInstance taskInstance) {
    super(taskInstance.getUUID());
    //this.candidatesInit = taskInstance.isCandidatesInitialized();
    try {
      this.initialUserId = taskInstance.getInitialTaskUser();
    } catch (IllegalStateException e) {
      this.initialUserId = null;
    }
    try {
      this.initialCandidates = Tool.copy(taskInstance.getInitialTaskCandidates());
    } catch (IllegalStateException e) {
      this.initialCandidates = null;
    }

    this.startedDate = taskInstance.getStartedDate();
    this.startedBy = taskInstance.getStartedBy();
    this.endedDate = taskInstance.getEndedDate();
    this.endedBy = taskInstance.getEndedBy();
    this.dueDate = taskInstance.getDueDate();
    this.createdDate = taskInstance.getCreatedDate();

    List<Update> list = taskInstance.getUpdates();
    if (list != null && !list.isEmpty()) {
      this.updates = new ArrayList<Update>();
      for (Update update : list) {
        if (update instanceof StateUpdateImpl) {
          this.updates.add(new StateUpdateImpl((StateUpdateImpl)update));
        } else if (update instanceof AssignUpdateImpl) {
          this.updates.add(new AssignUpdateImpl((AssignUpdateImpl)update));
        }
      }
    }
  }

  public ActivityBody copy() {
    return new TaskInstanceImpl(this);
  }

  public String toString() {
    String userId;
    try {
      userId = getTaskUser();
    } catch (IllegalStateException e) {
      userId = null;
    }
    Set<String> candidates;
    try {
      candidates = getTaskCandidates();
    } catch (IllegalStateException e) {
      candidates = null;
    }

    return this.getClass().getName()
      + "[taskUUID: " + getUUID()
      + ", initialUserId: " + initialUserId
      + ", initialCandidates: " + initialCandidates
      + ", userId: " + userId
      + ", candidates: " + candidates
      + ", state: " + getState()
      + ", createdDate: " + getCreatedDate()
      + ", dueDate: " + getDueDate()
      + ", startedBy: " + getStartedBy()
      + ", startedDate: " + getStartedDate()
      + ", endedDate: " + getEndedDate()
      + ", endedBy: " + getEndedBy()
      + "]";
  }

  public Date getCreatedDate() {
    return createdDate;
  }

  public Date getDueDate() {
    return dueDate;
  }

  public String getEndedBy() {
    return endedBy;
  }

  public String getStartedBy() {
    return startedBy;
  }

  public String getUpdatedBy() {
    if (updates != null && !updates.isEmpty()) {
      if (updates.get(updates.size() - 1).getUpdatedBy() != null) {
        return updates.get(updates.size() - 1).getUpdatedBy();
      }
      throw new IllegalStateException("The task: " + this.getUUID() + " has never been updated");
    }
    return null;
  }

  public String getTaskUser() {
    //only AssignUpdate type has recorded userId
    if (updates != null && !updates.isEmpty()) {
      int index = updates.size() - 1;
      while (index != 0) {
        if (updates.get(index) instanceof AssignUpdate) {
          AssignUpdate assignUpdate = (AssignUpdate) updates.get(index);
          if (assignUpdate.getUserId() != null) {
            return assignUpdate.getUserId();
          }
          throw new IllegalStateException("The task: " + this.getUUID() + " has not been assigned");
        }
        index--;
      }
    }
    return getInitialTaskUser();
  }

  public String getInitialTaskUser() {
    if (!isTaskInitiallyAssigned()) {
      throw new IllegalStateException("The task: " + this.getUUID() + " has not been initialy assigned");
    }
    return initialUserId;
  }

  public Set<String> getTaskCandidates() {
    if (updates != null && !updates.isEmpty()) {
      int index = updates.size() - 1;
      while (index != 0) {
        if (updates.get(index) instanceof AssignUpdate) {
          AssignUpdate assignUpdate = (AssignUpdate) updates.get(index);
          if (assignUpdate.getCandidates() != null) {
            return assignUpdate.getCandidates();
          }
        }
        index--;
      }
    }
    return getInitialTaskCandidates();
  }

  public Set<String> getInitialTaskCandidates() {
    return initialCandidates;
  }

  public TaskState getState() {
    if (updates != null && !updates.isEmpty()) {
      if (updates.get(updates.size() - 1).getTaskState() != null) {
        return updates.get(updates.size() - 1).getTaskState();
      }
      throw new IllegalStateException("The state of the task: " + this.getUUID() + " is null");
    }
    return initialState;
  }

  public List<Update> getUpdates() {
    return updates;
  }

  @Override
  public TaskUUID getUUID() {
    return (TaskUUID) super.getUUID();
  }

  public boolean isTaskInitiallyAssigned() {
    return initialUserId != null;
  }

  public boolean isTaskAssigned() {
    try {
      return getTaskUser() != null;
    } catch (IllegalStateException e) {
      return false;
    }
  }


  public TaskState getInitialState() {
    return initialState;
  }



}
