/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.parsing.binding;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.internal.util.XmlUtil;
import org.jbpm.pvm.internal.xml.Parse;
import org.jbpm.pvm.internal.xml.Parser;
import org.ow2.bonita.definition.VariablePerformerAssign;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.InstantiatorDefinition;
import org.ow2.bonita.facade.def.element.PerformerAssignDefinition;
import org.ow2.bonita.facade.def.element.SimulationInformationDefinition;
import org.ow2.bonita.facade.def.element.SubFlowDefinition;
import org.ow2.bonita.facade.def.element.TimeEstimationDefinition;
import org.ow2.bonita.facade.def.element.ToolDefinition;
import org.ow2.bonita.facade.def.element.TransitionRestrictionDefinition;
import org.ow2.bonita.facade.def.element.TransitionRestrictionDefinition.JoinType;
import org.ow2.bonita.facade.def.element.TransitionRestrictionDefinition.SplitType;
import org.ow2.bonita.facade.def.element.impl.DeadlineDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.HookDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.InstantiatorDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.PerformerAssignDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.SimulationInformationDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.SubFlowDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.TimeEstimationDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.TransitionRestrictionDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.impl.ActivityDefinitionImpl.IterationDefinitionImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.novabpm.util.Misc;
import org.w3c.dom.Element;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class ActivityBinding extends MajorElementBinding {

  public ActivityBinding() {
    super("Activity");
  }
  private static Logger log = Logger.getLogger(PackageBinding.class.getName());

  public Object parse(Element activityElement, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + activityElement);
    }

    String id = getId(activityElement);
    String name = getName(activityElement);

    ActivityDefinitionUUID recordUUID = new ActivityDefinitionUUID(
        getUUID(PackageDefinitionUUID.class, parse), getUUID(ProcessDefinitionUUID.class, parse), name);
    parse.pushObject(recordUUID);

    String description = getChildTextContent(activityElement, "Description");
    String limit = getChildTextContent(activityElement, "Limit");
    if (limit != null) {
      parse.addProblem("'Limit' element not yet supported on activity definition."
          + "Please remove it from activity: " + id);
      if ("".equals(limit.trim())) {
        parse.addProblem("Limit definition incorrect: no value specified in element 'Limit' for activity: " + id);
      }
    }
    boolean route = XmlUtil.element(activityElement, "Route") != null;
    boolean noImplementation = false;
    SubFlowDefinition subFlow = null;
    Set<ToolDefinition> tools = null;
    Element implementationElement = XmlUtil.element(activityElement, "Implementation");
    if (implementationElement != null) {
      noImplementation = XmlUtil.element(implementationElement, "No") != null;
      subFlow = getSubflowDef(implementationElement, parse);
      tools = getToolDef(implementationElement, parse);
    }
    String blockId = null;
    Element blockActivityElement = XmlUtil.element(activityElement, "BlockActivity");
    if (blockActivityElement != null) {
      blockId = XmlUtil.attribute(blockActivityElement, "Id");
    }

    String performer = getChildTextContent(activityElement, "Performer");

    ActivityDefinition.StartMode startMode = ActivityDefinition.StartMode.Automatic;
    Element startModeElement = XmlUtil.element(activityElement, "StartMode");
    if (startModeElement != null) {
      Element startModeFirstChildElement = XmlUtil.element(startModeElement);
      startMode = getEnumValue(ActivityDefinition.StartMode.class,
          startModeFirstChildElement.getLocalName(), ActivityDefinition.StartMode.Automatic);
    }
    ActivityDefinition.FinishMode finishMode = null;
    Element finishModeElement = XmlUtil.element(activityElement, "FinishMode");
    if (finishModeElement != null) {
      Element finishModeFirstChildElement = XmlUtil.element(finishModeElement);
      finishMode = getEnumValue(ActivityDefinition.FinishMode.class,
          finishModeFirstChildElement.getLocalName(), ActivityDefinition.FinishMode.Automatic);
    } else {
      if (startMode.equals(ActivityDefinition.StartMode.Automatic)) {
        finishMode = ActivityDefinition.FinishMode.Automatic;
      } else if (startMode.equals(ActivityDefinition.StartMode.Automatic)) {
        finishMode = ActivityDefinition.FinishMode.Automatic;
      } else {
        finishMode = ActivityDefinition.FinishMode.Manual;
      }
    }
    checkStartAndFinishMode(startMode, finishMode, performer, id, noImplementation,parse);
    String priority = getChildTextContent(activityElement, "Priority");
    if (priority != null) {
      parse.addProblem("'Priority' element not yet supported on activity definition. Please remove it from activity: " + id);
    }

    Set<DeadlineDefinition> deadlines = getDeadlines(activityElement, parse);
    SimulationInformationDefinition simulationInformation = getSimulationInformation(activityElement);

    String icon = getChildTextContent(activityElement, "Icon");
    String documentation = getChildTextContent(activityElement, "Documentation");

    Set<TransitionRestrictionDefinition> transitionRestrictions = null;
    Element transitionRestrictionsElement = XmlUtil.element(activityElement, "TransitionRestrictions");
    if (transitionRestrictionsElement != null) {
      List<Element> transitionRestrictionElements = XmlUtil.elements(transitionRestrictionsElement, "TransitionRestriction");
      if (transitionRestrictionElements != null) {
        transitionRestrictions = new HashSet<TransitionRestrictionDefinition>();
        for (Element transitionrestrictionElement : transitionRestrictionElements) {
          JoinType joinType = getJoinType(transitionrestrictionElement, parse);
          SplitType splitType = getSplitType(transitionrestrictionElement, parse);
          Set<String> splitTransitionRefIds = getSplitTransitionRefIds(transitionrestrictionElement, parse);
          TransitionRestrictionDefinition transitionRestrictionDefinition =
            new TransitionRestrictionDefinitionImpl(joinType, splitType, splitTransitionRefIds);
          transitionRestrictions.add(transitionRestrictionDefinition);
        }
      }
    }


    boolean isManualActivity =
      startMode.equals(ActivityDefinition.StartMode.Manual) || finishMode.equals(ActivityDefinition.FinishMode.Manual);
    boolean isAutomaticActivity =
      startMode.equals(ActivityDefinition.StartMode.Automatic) || finishMode.equals(ActivityDefinition.FinishMode.Automatic);

    final Set<HookDefinition> hooks = getHooks(activityElement, parse, isAutomaticActivity);

    if (isManualActivity && (route || (noImplementation && (subFlow != null || blockId != null)))) {
      parse.addProblem("Impossible to define a 'manual activity' (startMode or finishMode = Manual) "
          + "of one of the following types: route, noImplemention/subFlow, noImplementation/blockActivity");
    }

    if (isAutomaticActivity && hooks != null) {
      for (HookDefinition hookDefinition : hooks) {
        final HookDefinition.Event event = hookDefinition.getEvent();
        if (!HookDefinition.Event.automaticOnEnter.equals(event)) {
          parse.addProblem("Unsupported HookEventName: " + event + " for automatic activity with id: " + id
              + ". Hook class = " + hookDefinition.getClassName()
              + ". Valid HookEventName for automatic activity is: " + HookDefinition.Event.automaticOnEnter);
        }
      }
    }
    if (isManualActivity && hooks != null) {
      for (HookDefinition hook : hooks) {
        if (!HookDefinition.Event.TASK_EVENTS.contains(hook.getEvent())) {
          parse.addProblem("Task activities (NoImplementation - manual) can only define hook on event: "
              + HookDefinition.Event.TASK_EVENTS + ". Hook class = "
              + hook.getClassName());
        }
      }
    }
    PerformerAssignDefinition performerAssign = getPerformerAssign(activityElement, parse);
    Set<DataFieldDefinition> dataFields = getDataFields(activityElement, parse);

    Set<Element> iterationElements = getExtendedAttributes(activityElement, "Iteration");
    Set<ActivityDefinition.IterationDef> iterations = null;
    if (iterationElements != null) {
      iterations = new HashSet<ActivityDefinition.IterationDef>();
      for (Element iterationElement : iterationElements) {
        String iterationCondition = XmlUtil.attribute(iterationElement, "Value");
        String iterationTo = getChildTextContent(iterationElement, "To");
        iterations.add(new IterationDefinitionImpl(iterationTo, iterationCondition));
      }
    }

    InstantiatorDefinition activityInstantiator = null;
    Element activityInstantiatorElement = getExtendedAttribute(activityElement, "MultiInstantiation");
    if (activityInstantiatorElement != null) {
      activityInstantiator = new InstantiatorDefinitionImpl(XmlUtil.attribute(activityInstantiatorElement, "Value"), null);
    }

    ActivityDefinition activityDefinition = new ActivityDefinitionImpl(recordUUID, id,
        getUUID(PackageDefinitionUUID.class, parse), getUUID(ProcessDefinitionUUID.class, parse),
        blockId, deadlines, description, documentation, finishMode,
        icon, limit, name, performer, priority, simulationInformation,
        startMode, subFlow, tools, transitionRestrictions,
        noImplementation, route, hooks, performerAssign, dataFields,
        iterations, activityInstantiator
    );
    parse.popObject();
    return activityDefinition;
  }

  private JoinType getJoinType(Element transitionrestrictionElement, Parse parse) {
    JoinType joinType = null;
    Element joinElement = XmlUtil.element(transitionrestrictionElement, "Join");
    if (joinElement != null) {
      joinType = getEnumValue(JoinType.class, XmlUtil.attribute(joinElement, "Type"), null);
      if (joinType == null) {
        parse.addProblem("Mandatory type attribute is not specified on join element");
      }
    }
    return joinType;
  }

  private SplitType getSplitType(Element transitionrestrictionElement, Parse parse) {
    SplitType splitType = null;
    Element splitElement = XmlUtil.element(transitionrestrictionElement, "Split");
    if (splitElement != null) {
      splitType = getEnumValue(SplitType.class, XmlUtil.attribute(splitElement, "Type"), null);
      if (splitType == null) {
        parse.addProblem("Mandatory type attribute is not specified on Split element!");
      }
    }
    return splitType;
  }
  private Set<String> getSplitTransitionRefIds(Element transitionrestrictionElement, Parse parse) {
    Set<String> splitTransitionRefIds = null;
    Element splitElement = XmlUtil.element(transitionrestrictionElement, "Split");
    if (splitElement != null) {
      Element transitionRefsElement = XmlUtil.element(transitionrestrictionElement, "TransitionRefs");
      if (transitionRefsElement != null) {
        parse.addProblem("TransitionRefs element on Split is not supported.");
        splitTransitionRefIds = new HashSet<String>();
        List<Element> transitionRefElements = XmlUtil.elements(transitionRefsElement, "TransitionRef");
        for (Element transitionRefElement : transitionRefElements) {
          String id = XmlUtil.attribute(transitionRefElement, "Id");
          splitTransitionRefIds.add(id);
        }
      }
    }
    return splitTransitionRefIds;
  }


  protected PerformerAssignDefinition getPerformerAssign(Element activityElement, Parse parse) {
    Element performerAssignElement = getExtendedAttribute(activityElement, "PerformerAssign");
    if (performerAssignElement != null) {
      String value = XmlUtil.attribute(performerAssignElement, "Value");
      String className = null;
      Map<String, String> parameters = null;
      if ("Callback".equals(value)) {
        Element callbackElement = XmlUtil.element(performerAssignElement, "Callback");
        className = callbackElement.getTextContent();
      } else if ("Custom".equals(value)) {
        Element callbackElement = XmlUtil.element(performerAssignElement, "Custom");
        className = callbackElement.getTextContent();
      } else if ("Variable".equals(value)) {
        Element propertyElement = XmlUtil.element(performerAssignElement, "Variable");
        parameters = new HashMap<String, String>();
        parameters.put("variableId", propertyElement.getTextContent());
        className = VariablePerformerAssign.class.getName();
      } else if ("Property".equals(value)) { // Bonita/ProEd V3
        parse.addWarning("PerformerAssign: " + value
            + ". \"Property\" type is a Bonita v3 type name. Bonita v4 type name is \"Variable\". Class used will be: "
            + VariablePerformerAssign.class.getName());
        Element propertyElement = XmlUtil.element(performerAssignElement, "Property");
        parameters = new HashMap<String, String>();
        parameters.put("variableId", propertyElement.getTextContent());
        className = VariablePerformerAssign.class.getName();
      } else {
        parse.addProblem("Unsupported value on extendedAttribute PerformerAssign: " + value);
      }
      return new PerformerAssignDefinitionImpl(className, parameters);
    }
    return null;
  }

  private void checkStartAndFinishMode(ActivityDefinition.StartMode startMode, ActivityDefinition.FinishMode finishMode,
      String performer, String activityId, boolean isNoImpl, Parse parse) {
    if (
        (ActivityDefinition.FinishMode.Manual.equals(finishMode) && ActivityDefinition.StartMode.Automatic.equals(startMode))
        ||
        (ActivityDefinition.FinishMode.Automatic.equals(finishMode) && ActivityDefinition.StartMode.Manual.equals(startMode))
    ) {
      parse.addProblem("StartMode and FinishMode have different values: this feature is not yet supported.");
    }
    boolean hasManualMode =
      ActivityDefinition.StartMode.Manual.equals(startMode) || ActivityDefinition.FinishMode.Manual.equals(finishMode);
    if (hasManualMode) {
      if (performer == null) {
        parse.addProblem("StartMode or FinishMode is Manual and no performer is specified on activity processDefinitionUUID = "
            + activityId + "! Please specify one.");
      }
      if (!isNoImpl) {
        parse.addProblem("StartMode or FinishMode is Manual and activity implementation is not No:"
            + "this feature is not yet supported.");
      }
    }
  }

  protected SimulationInformationDefinition getSimulationInformation(Element activityElement) {
    SimulationInformationDefinition simulationInformationDefinition = null;
    Element simulationInformationElement = XmlUtil.element(activityElement, "SimulationInformation");
    if (simulationInformationElement != null) {
      SimulationInformationDefinition.Instantiation instantiation =
        getEnumValue(SimulationInformationDefinition.Instantiation.class,
            XmlUtil.attribute(simulationInformationElement, "Instantiation"), null);
      String cost = getChildTextContent(simulationInformationElement, "Cost");
      TimeEstimationDefinition timeEstimation = null;
      Element timeEstimationElement = XmlUtil.element(simulationInformationElement, "TimeEstimation");
      if (timeEstimationElement != null) {
        String waitingTime = getChildTextContent(timeEstimationElement, "WaitingTime");
        String workingTime = getChildTextContent(timeEstimationElement, "WorkingTime");
        String duration = getChildTextContent(timeEstimationElement, "Duration");
        timeEstimation = new TimeEstimationDefinitionImpl(duration, waitingTime, workingTime);
      }
      simulationInformationDefinition = new SimulationInformationDefinitionImpl(cost, instantiation, timeEstimation);
    }
    return simulationInformationDefinition;
  }

  protected Set<ToolDefinition> getToolDef(Element implementationElement, Parse parse) {
    Element toolElement = XmlUtil.element(implementationElement, "Tool");
    if (toolElement != null) {
      parse.addProblem("Tool is not supported as an implementationType");
      //implementation.setImplementationType(activity.new ToolImplementation());
    }
    return null;
  }

  protected SubFlowDefinition getSubflowDef(Element implementationElement, Parse parse) {
    SubFlowDefinition subFlowDefinition = null;
    Element subFlowElement = XmlUtil.element(implementationElement, "SubFlow");
    if (subFlowElement != null) {
      List<String> actualParameters = null;
      Element actualparametersElement = XmlUtil.element(subFlowElement, "ActualParameters");
      if (actualparametersElement != null) {
        List<Element> actualparameterElements = XmlUtil.elements(actualparametersElement, "ActualParameter");
        if (actualparameterElements != null) {
          actualParameters = new ArrayList<String>();
          for (Element actualparameterElement : actualparameterElements) {
            actualParameters.add(actualparameterElement.getTextContent());
          }
        }
      }

      String id = XmlUtil.attribute(subFlowElement, "Id");
      SubFlowDefinition.Execution execution = getEnumValue(SubFlowDefinition.Execution.class,
          XmlUtil.attribute(subFlowElement, "Execution"), null);
      if (execution != null) {
        if (SubFlowDefinition.Execution.ASYNCHR.equals(execution)) {
          parse.addProblem("Asyncr SubFlow is not supported");
        }
      }
      subFlowDefinition = new SubFlowDefinitionImpl(actualParameters, execution, id);
    }
    return subFlowDefinition;
  }

  @SuppressWarnings("unchecked")
  protected Set<DataFieldDefinition> getDataFields(Element activityElement, Parse parse) {
    Set<DataFieldDefinition> result = null;
    Set<Element> propertyElements = getExtendedAttributes(activityElement, "property");
    if (propertyElements != null) {
      Set<DataFieldDefinition> activityDataFields = new HashSet<DataFieldDefinition>();
      activityDataFields = parse.findObject(activityDataFields.getClass());
      if (activityDataFields == null) {
        parse.addProblem("No activity dataFields is defined within the process.");
      } else {
        result = new HashSet<DataFieldDefinition>();
        for (Element propertyElement : propertyElements) {
          String dataFieldId = XmlUtil.attribute(propertyElement, "Value");
          DataFieldDefinition dataField = getActivityDataField(activityDataFields, dataFieldId);
          if (dataField == null) {
            parse.addProblem("Looking for a datafield with processDefinitionUUID: " + dataFieldId
                + " in enclosing process but unable to find it.");
          } else {
            result.add(dataField);
          }
          String propagatedValue = getChildTextContent(propertyElement, "Propagated");
          if (propagatedValue != null && !("no".equalsIgnoreCase(propagatedValue) || "false".equalsIgnoreCase(propagatedValue))) {
            parse.addProblem("Propagated value not supported: " + propagatedValue
                + ". Use instance variables instead."
                + "(Only 'no' or 'false' are supported for backward compatibility.)");
          }
        }
      }
    }
    return result;
  }

  protected DataFieldDefinition getActivityDataField(Set<DataFieldDefinition> activityDataFields, String dataFieldId) {
    for (DataFieldDefinition dataField : activityDataFields) {
      if (dataField.getDataFieldId().equals(dataFieldId)) {
        return dataField;
      }
    }
    return null;
  }
  protected Set<HookDefinition> getHooks(Element activityElement, Parse parse, boolean isAutomatic) {
    Set<Element> hookElements = getExtendedAttributes(activityElement, "hook");
    Set<HookDefinition> hooks = null;
    if (hookElements != null) {
      hooks = new HashSet<HookDefinition>();
      for (Element hookElement : hookElements) {
        String className = XmlUtil.attribute(hookElement, "Value");
        String hookEventName = getChildTextContent(hookElement, "HookEventName");
        if (hookEventName == null) {
          parse.addProblem("hook ExtendedAttribute needs an element child called HookEventName");
        } else {
          String rollBackFlag = getChildTextContent(hookElement, "Rollback");
          if (rollBackFlag == null) {
            parse.addWarning("Rollback flag is not specified for hook: " + hookEventName
                + ". Using Bonita v3 hook compatibility parser.");
            //beforeTerminate, afterTerminate, beforeFinish, afterFinish,
            //beforeStart, afterStart, onDeadline, taskOnCancel, taskOnReady
            HookDefinition.V3Event v3Event = null;
            try {
              v3Event = Misc.stringToEnum(HookDefinition.V3Event.class, hookEventName);
            } catch (IllegalArgumentException iae) {
              parse.addProblem("Unsupported HookEventName: " + hookEventName + " for "
                  + (isAutomatic ? "automatic" : "manual") + " activity.");
            }
            if (v3Event != null) {
              final HookDefinition.Event event = HookDefinition.V3Event.getV4FromV3HookEventName(v3Event, isAutomatic);
              if (event == null) {
                parse.addProblem("Unsupported HookEventName: " + hookEventName + " for "
                    + (isAutomatic ? "automatic" : "manual") + " activity");
              } else {
                parse.addWarning("Bonita v3 HookEventName: " + v3Event + " is mapped to Bonita v4 HookEventName: "
                    + event + "' for " + (isAutomatic ? "automatic " : "manual") + " activity");
                boolean throwingException = ActivityBinding.getExceptionFlagFromV3Hook(v3Event);
                hooks.add(new HookDefinitionImpl(className, event, null, throwingException));
              }
            }
          } else { // Bonita v4 Hook Specification (see ProEd v4)
            HookDefinition.Event event = null;
            // We can't use Boolean's class methods since in their case,
            // true == "true" and false == !true (anything different to "true")
            // In our case, we really want either "true" or "false". Any other string
            // is simply an invalid flag.
            boolean throwingException = false;
            if ("true".equals(rollBackFlag)) {
              throwingException = true;
            } else if ("false".equals(rollBackFlag)) {
              throwingException = false;
            } else {
              parse.addProblem("Invalid Rollback specification: " + rollBackFlag + ". It should be either 'true' or 'false'.");
            }
            try {
              event = Misc.stringToEnum(HookDefinition.Event.class, hookEventName);
            } catch (IllegalArgumentException iae) {
              parse.addProblem("Unsupported HookEventName: " + hookEventName + " for "
                  + (isAutomatic ? "automatic" : "manual") + " activity.");
            }
            if (event != null) {
              hooks.add(new HookDefinitionImpl(className, event, null, throwingException));
            }
          }
        }
      }
    }
    return hooks;
  }

  private static boolean getExceptionFlagFromV3Hook(final HookDefinition.V3Event event) {
    if (event.equals(HookDefinition.V3Event.afterTerminate) || event.equals(HookDefinition.V3Event.beforeStart)) {
      return false;
    }
    return true;
  }

  protected Set<DeadlineDefinition> getDeadlines(Element activityElement, Parse parse) {
    List<Element> deadlineElements = XmlUtil.elements(activityElement, "Deadline");
    Set<DeadlineDefinition> deadlines = null;

    if (deadlineElements != null) {
      log.warning("Bonita use of deadlines is specific: refer to bonita User Guide for more details");
      deadlines = new HashSet<DeadlineDefinition>();
      for (Element deadlineElement : deadlineElements) {
        DeadlineDefinition.Execution execution =
          getEnumValue(DeadlineDefinition.Execution.class,
              XmlUtil.attribute(deadlineElement, "Execution"), null);
        if (execution == null) {
          execution = DeadlineDefinition.Execution.SYNCHR;
        }

        // schema validation ensures that deadlineCondtion and ExceptionName
        // are specified once and only once
        String deadlineCondition = getChildTextContent(deadlineElement, "DeadlineCondition");
        if (deadlineCondition == null) {
          parse.addProblem("DeadlineCondition element is not specified on deadline element");
        }

        String exceptionName =  getChildTextContent(deadlineElement, "ExceptionName");
        if (exceptionName == null) {
          parse.addProblem("ExceptionName element is not specified on deadline element");
        }

        deadlines.add(new DeadlineDefinitionImpl(deadlineCondition, exceptionName, execution));
      }
    }
    return deadlines;
  }
}


