/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * deploymentId 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.parsing.binding;

import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.internal.util.XmlUtil;
import org.jbpm.pvm.internal.xml.Parse;
import org.jbpm.pvm.internal.xml.Parser;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.facade.def.dataType.DataTypeDefinition;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.ExternalPackageDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.TypeDeclarationDefinition;
import org.ow2.bonita.facade.def.element.impl.ExternalPackageDefinitionImpl;
import org.ow2.bonita.facade.def.element.impl.TypeDeclarationDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivitySetDefinition;
import org.ow2.bonita.facade.def.majorElement.ApplicationDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.PackageFullDefinitionImpl;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.parsing.XpdlParserException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.XmlConstants;
import org.w3c.dom.Element;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */

public class PackageBinding extends MajorElementBinding {

  private static Logger log = Logger.getLogger(PackageBinding.class.getName());

  public PackageBinding() {
    super("Package");
  }
  
  static class PackageVersion {
    private String version;
    public PackageVersion(String version) {
      super();
      this.version = version;
    }
    public String getVersion() {
      return version;
    }
  }
  public Object parse(Element packageElement, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + packageElement);
    }

    String id = getId(packageElement);
    String name = getName(packageElement);

    Element packageHeaderElement = XmlUtil.element(packageElement, "PackageHeader");
    String xpdlVersion = getChildTextContent(packageHeaderElement, "XPDLVersion");
    String vendor = getChildTextContent(packageHeaderElement, "Vendor");
    String created = getChildTextContent(packageHeaderElement, "Created");
    String description = getChildTextContent(packageHeaderElement, "Description");
    String documentation = getChildTextContent(packageHeaderElement, "Documentation");
    String priorityUnit = getChildTextContent(packageHeaderElement, "PriorityUnit");
    String costUnit = getChildTextContent(packageHeaderElement, "CostUnit");

    String packageNS = packageElement.getNamespaceURI();
    if (xpdlVersion == null || "".equals(xpdlVersion)) {
      parse.addProblem("XPDLVersion attribute must be specified at package level.");
    } else if (!"1.0".equals(xpdlVersion)) {
      parse.addProblem("Unsupported XPDLVersion : " + xpdlVersion);
    } else if ("1.0".equals(xpdlVersion) && !packageNS.equals(XmlConstants.XPDL_1_0_NS)) {
      String incoherentValuesMessage =
        "XPDLVersion is " + xpdlVersion + ", and package namespace is " + packageNS + ". These two values are not coherents.";
      parse.addProblem(incoherentValuesMessage);
    }

    Element redefinableHeaderElement = XmlUtil.element(packageElement, "RedefinableHeader");
    String author = getChildTextContent(redefinableHeaderElement, "Author");
    String version = getChildTextContent(redefinableHeaderElement, "Version");
    if (version == null || "".equals(version)) {
      parse.addProblem("Version attribut is mandatory at Package level (in RedefinableHeader). Package " 
          + id + " does not define this attribut.");
    }
    String codepage = getChildTextContent(redefinableHeaderElement, "Codepage");
    String countrykey = getChildTextContent(redefinableHeaderElement, "Countrykey");
    String pubStatusAttr = null;

    PackageDefinitionUUID recordUUID = new PackageDefinitionUUID(name, version);
    parse.pushObject(recordUUID);

    if (redefinableHeaderElement != null) {
      pubStatusAttr = XmlUtil.attribute(redefinableHeaderElement, "PublicationStatus");
    }
    PackageDefinition.PublicationStatus publicationStatus =
      getEnumValue(PackageDefinition.PublicationStatus.class, pubStatusAttr, null);
    Set<String> responsibles = null;

    if (redefinableHeaderElement != null) {
      Element responsiblesElement = XmlUtil.element(redefinableHeaderElement, "Responsibles");
      if (responsiblesElement != null) {
        List<Element> responsibleElements = XmlUtil.elements(responsiblesElement, "Responsible");
        responsibles = new HashSet<String>();
        for (Element responsibleElement : responsibleElements) {
          responsibles.add(responsibleElement.getTextContent());
        }
      }
    }
    PackageDefinition.GraphConformance conformanceClass = parseConformanceClass(packageElement, parse);
    Element scriptElement = XmlUtil.element(packageElement, "Script");

    String scriptType = "bsh";
    String scriptVersion = "2.0b1";
    String scriptGrammar = null;
    if (scriptElement != null) {
      scriptGrammar = XmlUtil.attribute(scriptElement, "Grammar");
      scriptType = XmlUtil.attribute(scriptElement, "Type");
      scriptVersion = XmlUtil.attribute(scriptElement, "Version");
      if (!"bsh".equals(scriptType) || !"2.0b1".equals(scriptVersion)) {
        parse.addProblem("Unsupported script : " + scriptType + "-" + scriptVersion + "-" + scriptGrammar);
      }
    }


    parse.pushObject(new PackageVersion(version));
    Set<ExternalPackageDefinition> externalPackages = parseExternalPackages(packageElement, parse);
    Set<TypeDeclarationDefinition> typeDeclarations = parseTypeDeclarations(packageElement, parse, parser);

    Set<ParticipantDefinition> participants = parseXpdlMajorElementList(packageElement, 
        "Participants", "Participant", parse, parser);
    Set<ApplicationDefinition> applications = parseXpdlMajorElementList(packageElement, 
        "Applications", "Application", parse, parser);
    Set<DataFieldDefinition> dataFields = parseXpdlMajorElementList(packageElement,"DataFields", "DataField", parse, parser);
    Set<ProcessFullDefinition> processes = parseXpdlMajorElementList(packageElement, 
        "WorkflowProcesses", "WorkflowProcess", parse, parser);

    Set<String> classDependencies = getClassDependencies(processes, participants);
    Set<String> packageDependencies = getPackageDependencies(processes);

    Date deployedDate = new Date();
    String deployedBy = EnvTool.getUserId();

    PackageDefinition packageDefinition = new PackageFullDefinitionImpl(recordUUID, deployedDate, deployedBy,
        id, name, author, classDependencies, codepage, conformanceClass, costUnit,
        countrykey, created, description,documentation, externalPackages, packageDependencies, priorityUnit,
        publicationStatus, responsibles, scriptType, scriptVersion, scriptGrammar,
        typeDeclarations, vendor, version, xpdlVersion, participants, applications, dataFields, processes);

    //pop packageVersion
    parse.popObject();
    
    parse.popObject();
    return packageDefinition;
  }

  protected Set<String> getPackageDependencies(Set<ProcessFullDefinition> processes) {
    Set<String> dependencies = new HashSet<String>();
    if (processes != null) {
      for (ProcessFullDefinition process : processes) {
        if (process.getActivities() != null) {
          for (ActivityDefinition activity : process.getActivities()) {
            if (activity.getSubFlow() != null) {
              String subProcessId = activity.getSubFlow().getProcessId();
              if (!containsProcess(processes, subProcessId)) {
                ProcessFullDefinition subProcess =
                  EnvTool.getJournalQueriers().getLastDeployedProcess(activity.getSubFlow().getProcessId());

                if (subProcess == null) {
                  throw new DeploymentRuntimeException("Unable to find process \"" + subProcessId + "\" used in subFlow. "
                      + "If the subFlow process is in an external package, "
                      + "please check that this package has already been deployed");
                }
                PackageDefinition packageDefinition = 
                  EnvTool.getJournalQueriers().getPackage(subProcess.getPackageDefinitionUUID());
                if (packageDefinition.getUndeployedBy() != null) {
                  throw new DeploymentRuntimeException("Unable to find process \"" + subProcessId + "\" used in subFlow. ");
                }
                dependencies.add(packageDefinition.getUUID().toString());
              }
            }
          }
        }
        if (process.getActivitySets() != null) {
          for (ActivitySetDefinition activitySet : process.getActivitySets()) {
            dependencies.addAll(getActivitiesDependencies(activitySet.getActivities()));
          }
        }
      }
    }
    if (dependencies.isEmpty()) {
      dependencies = null;
    }
    return dependencies;
  }

  protected boolean containsProcess(Set<ProcessFullDefinition> processes, String processId) {
    if (processes == null) {
      return false;
    }
    for (ProcessFullDefinition process : processes) {
      if (process.getProcessId().equals(processId)) {
        return true;
      }
    }
    return false;
  }

  protected Set<String> getClassDependencies(Set<ProcessFullDefinition> processes,
      Set<ParticipantDefinition> participants) {
    Set<String> dependencies = new HashSet<String>();
    dependencies.addAll(getParticipantsDependencies(participants));
    if (processes != null) {
      for (ProcessFullDefinition process : processes) {
        dependencies.addAll(getActivitiesDependencies(process.getActivities()));
        dependencies.addAll(getParticipantsDependencies(process.getParticipants()));
        if (process.getActivitySets() != null) {
          for (ActivitySetDefinition activitySet : process.getActivitySets()) {
            dependencies.addAll(getActivitiesDependencies(activitySet.getActivities()));
          }
        }
      }
    }
    if (dependencies.isEmpty()) {
      dependencies = null;
    }
    return dependencies;
  }
  protected Set<String> getActivitiesDependencies(Set<ActivityDefinition> activities) {
    Set<String> dependencies = new HashSet<String>();
    if (activities != null) {
      for (ActivityDefinition activity : activities) {
        if (activity.getPerformerAssign() != null) {
          dependencies.add(activity.getPerformerAssign().getClassName());
        }
        if (activity.getHooks() != null) {
          for (HookDefinition hook : activity.getHooks()) {
            dependencies.add(hook.getClassName());
          }
        }
        if (activity.getDeadlines() != null) {
          for (DeadlineDefinition deadline : activity.getDeadlines()) {
            dependencies.add(deadline.getExceptionName());
          }
        }
        if (activity.getActivityInstantiator() != null) {
          dependencies.add(activity.getActivityInstantiator().getClassName());
        }
      }
    }
    return dependencies;
  }
  protected Set<String> getParticipantsDependencies(Set<ParticipantDefinition> participants) {
    Set<String> dependencies = new HashSet<String>();
    if (participants != null) {
      for (ParticipantDefinition participant : participants) {
        if (participant.getRoleMapper() != null) {
          dependencies.add(participant.getRoleMapper().getClassName());
        }
      }
    }
    return dependencies;
  }

  protected PackageDefinition.GraphConformance parseConformanceClass(Element packageElement, Parse parse) {
    Element conformanceClassElement = XmlUtil.element(packageElement, "ConformanceClass");
    if (conformanceClassElement != null) {
      PackageDefinition.GraphConformance conformanceClass = getEnumValue(PackageDefinition.GraphConformance.class,
          XmlUtil.attribute(conformanceClassElement, "GraphConformance"), null);
      if (conformanceClass == null) {
        parse.addProblem("ConformanceClass element is defined on package but the attribute GraphConformance is not specified.");
      } else {
        if (!PackageDefinition.GraphConformance.NON_BLOCKED.equals(conformanceClass)) {
          parse.addProblem("Only " + PackageDefinition.GraphConformance.NON_BLOCKED
              + " value is supported for conformanceClass on package element.");
        } else {
          return conformanceClass;
        }
      }
    }
    return null;
  }

  protected Set<ExternalPackageDefinition> parseExternalPackages(Element packageElement, Parse parse) {
    Element externalPackagesElement = XmlUtil.element(packageElement, "ExternalPackages");
    Set<ExternalPackageDefinition> externalPackages = null;
    if (externalPackagesElement != null) {
      List<Element> externalPackageElements = XmlUtil.elements(externalPackagesElement, "ExternalPackage");
      if (externalPackageElements != null) {
        parse.addProblem("ExternalPackage element not yet supported.");
        externalPackages = new HashSet<ExternalPackageDefinition>();
        for (Element externalPackageElement : externalPackageElements) {
          String href = XmlUtil.attribute(externalPackageElement, "href");
          ExternalPackageDefinition externalPackageDefinition = new ExternalPackageDefinitionImpl(href);
          externalPackages.add(externalPackageDefinition);
        }
      }
    }
    return externalPackages;
  }

  protected Set<TypeDeclarationDefinition> parseTypeDeclarations(Element packageElement,
      Parse parse, Parser parser) {
    Element typeDeclarationsElement = XmlUtil.element(packageElement, "TypeDeclarations");
    Set<TypeDeclarationDefinition> typeDeclarations = null;
    if (typeDeclarationsElement != null) {
      List<Element> typeDeclarationElements = XmlUtil.elements(typeDeclarationsElement, "TypeDeclaration");
      if (typeDeclarationElements != null) {
        typeDeclarations = new HashSet<TypeDeclarationDefinition>();
        Set<String> ids = new HashSet<String>();
        Set<String> names = new HashSet<String>();
        for (Element typeDeclarationElement : typeDeclarationElements) {
          Element dataTypesElement = XmlUtil.element(typeDeclarationElement, "DataTypes");
          Set<DataTypeDefinition> dataTypes = null;
          if (dataTypesElement != null) {
            List<Element> dataTypeElements = XmlUtil.elements(dataTypesElement);
            if (dataTypeElements != null) {
              dataTypes = new HashSet<DataTypeDefinition>();
              for (Element dataTypeElement : dataTypeElements) {
                DataTypeDefinition dataType = (DataTypeDefinition) parser.parseElement(dataTypeElement, parse, "dataTypes");
                dataTypes.add(dataType);
              }
            }
          }
          String description = getChildTextContent(typeDeclarationElement, "Description");
          String id = XmlUtil.attribute(typeDeclarationElement, "Id");
          String name = XmlUtil.attribute(typeDeclarationElement, "Name");
          if (name == null) {
            name = id;
          }
          if (!ids.add(id)) {
            throw new XpdlParserException("Error: Duplicate packageDefinitionUUID " + id + " for typeDeclaration");
          }
          if (!names.add(name)) {
            throw new XpdlParserException("Error: Duplicate name " + name + " for typeDeclaration");
          }
          TypeDeclarationDefinition typeDeclarationDefinition = new TypeDeclarationDefinitionImpl(dataTypes, description,
              id, name);
          typeDeclarations.add(typeDeclarationDefinition);
        }
      }
    }
    return typeDeclarations;
  }

}
