package org.ow2.bonita.persistence.db;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.job.Timer;
import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PackageDependency;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.persistence.XpdlPersistenceService;
import org.ow2.bonita.runtime.ActivityInstanceKey;
import org.ow2.bonita.runtime.ProcessInstanceKey;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.util.EnvTool;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class DbRepository implements Repository {

  private static final class BufferRepository {
    private BufferRepository() { }
    private Map<ProcessInstanceUUID, XpdlInstance> bufferedXpdlInstances =
      new HashMap<ProcessInstanceUUID, XpdlInstance>();

    private Set<XpdlInstance> removedInstances = new HashSet<XpdlInstance>();

    private Map<TaskUUID, TaskRunTime> bufferedTasks =
      new HashMap<TaskUUID, TaskRunTime>();

    private Set<TaskRunTime> removedTasks = new HashSet<TaskRunTime>();
    
    private Map<PackageDefinitionUUID, PackageClassData> bufferedPackageClassData = 
      new HashMap<PackageDefinitionUUID, PackageClassData>();

    private Set<String> removedGlobalClassName = new HashSet<String>();
  }

  private BufferRepository buffer = new BufferRepository();

  private String persistenceServiceName;
  
  private XpdlPersistenceService getPersistenceService() {
    return EnvTool.getPersistenceService(persistenceServiceName);
  }
  
  public DbRepository(String persistenceServiceName) {
    this.persistenceServiceName = persistenceServiceName;
  }
  
  /*************************
   ** INSTANCE REPOSITORY **
   *************************/
  public XpdlInstance getXpdlInstance(ProcessInstanceUUID instanceUUID) {
    if (buffer.bufferedXpdlInstances.containsKey(instanceUUID)) {
      return buffer.bufferedXpdlInstances.get(instanceUUID);
    }
    XpdlInstance instance = getPersistenceService().getXpdlInstance(instanceUUID);
    if (instance != null) {
      buffer.bufferedXpdlInstances.put(instance.getUUID(), instance);
    }
    return instance;
  }

  public Set<XpdlInstance> getXpdlInstances(ProcessDefinitionUUID processId) {
    Set<XpdlInstance> result = new HashSet<XpdlInstance>();
    for (XpdlInstance instance : buffer.bufferedXpdlInstances.values()) {
      if (instance.getProcessDefinitionUUID().equals(processId)) {
        result.add(instance);
      }
    }
    result.addAll(getPersistenceService().getXpdlInstances(processId));
    result.removeAll(buffer.removedInstances);
    return result;
  }

  public void storeInstance(XpdlInstance instance) {
    getPersistenceService().save(instance);
    buffer.bufferedXpdlInstances.put(instance.getUUID(), instance);
  }

  public XpdlInstance removeXpdlInstance(ProcessInstanceUUID instanceUUID) {
    XpdlInstance instance = getXpdlInstance(instanceUUID);
    if (instance == null) {
      return null;
    }
    buffer.bufferedXpdlInstances.remove(instanceUUID);
    buffer.removedInstances.add(instance);
    getPersistenceService().delete(instance);
    return instance;
  }

  private XpdlExecution getExecOnNode(XpdlExecution exec, String nodeName) {
    if (exec.getExecutions() == null || exec.getExecutions().isEmpty()) {
      if (exec.getNode() != null && exec.getNode().getName().equals(nodeName)) {
        return exec;
      }
    }
    for (Execution child : exec.getExecutions()) {
      XpdlExecution found = getExecOnNode((XpdlExecution) child, nodeName);
      if (found != null) {
        return found;
      }
    }
    return null;
  }

  public XpdlExecution getExecutionPointingOnNode(ProcessInstanceUUID instanceUUID, String nodeName) {
    if (buffer.bufferedXpdlInstances.containsKey(instanceUUID)) {
      XpdlInstance instance = buffer.bufferedXpdlInstances.get(instanceUUID);
      return getExecOnNode(instance.getRootExecution(), nodeName);
    }
    XpdlExecution exec = getPersistenceService().getExecutionPointingOnNode(instanceUUID, nodeName);
    if (exec != null && !buffer.removedInstances.contains(exec.getXpdlInstance())) {
      return exec;
    }
    return null;
  }

  /************************
   ** PROCESS REPOSITORY **
   ************************/

  public XpdlProcess findXpdlProcessById(ProcessDefinitionUUID processId) {
    return getPersistenceService().findXpdlProcessById(processId);
  }

  public List<XpdlProcess> getXpdlProcesses() {
    return getPersistenceService().getXpdlProcesses();
  }

  public void storeProcess(XpdlProcess xpdlProcess) {
    getPersistenceService().save(xpdlProcess);
  }

  public XpdlProcess removeProcess(ProcessDefinitionUUID processId) {
    XpdlProcess p = findXpdlProcessById(processId);
    if (p == null) {
      return null;
    }
    getPersistenceService().delete(p);
    return p;
  }

  public boolean hasProcessWithUUID(ProcessDefinitionUUID uuid) {
    return getPersistenceService().hasProcessWithUUID(uuid);
  }

  public boolean hasPackageWithUUID(PackageDefinitionUUID packageUUID) {
    return getPersistenceService().hasPackageWithUUID(packageUUID);
  }

  /*********************
   ** TASK REPOSITORY **
   *********************/

  public void addTask(TaskRunTime task) {
    getPersistenceService().save(task);
  }

  public TaskRunTime getTask(TaskUUID taskUUID) {
    if (buffer.bufferedTasks.containsKey(taskUUID)) {
      return buffer.bufferedTasks.get(taskUUID);
    }
    TaskRunTime task = getPersistenceService().getTask(taskUUID);
    if (task != null && !buffer.removedTasks.contains(task)) {
      buffer.bufferedTasks.put(task.getUUID(), task);
    }
    return task;
  }

  public List<TaskRunTime> getInstanceTasks(ProcessInstanceUUID instanceUUID) {
    List<TaskRunTime> tasks = getPersistenceService().getInstanceTasks(instanceUUID);
    tasks.removeAll(buffer.removedTasks);
    return tasks;
  }

  public void removeTask(TaskRunTime task) {
    getPersistenceService().delete(task);
    buffer.removedTasks.add(task);
    buffer.bufferedTasks.remove(task.getUUID());
  }



  /*********************
   ** HOOK REPOSITORY **
   *********************/
  public PackageClassData getPackageClassData(PackageDefinitionUUID packageId) {
    if (buffer.bufferedPackageClassData.containsKey(packageId)) {
      return buffer.bufferedPackageClassData.get(packageId);
    }
    return getPersistenceService().getPackageClassData(packageId);
  }
  public void removePackageClassData(PackageDefinitionUUID packageId) {
    PackageClassData pcd = getPackageClassData(packageId);
    if (pcd == null) {
      return;
    }
    getPersistenceService().delete(pcd);
  }
  public void storePackageClassData(PackageClassData packageClassData) {
    getPersistenceService().save(packageClassData);
    buffer.bufferedPackageClassData.put(packageClassData.getPackageUUID(), packageClassData);
  }
  public GlobalClassData getGlobalClassData(String className) {
    if (buffer.removedGlobalClassName.contains(className)) {
      return null;
    }
    return getPersistenceService().getGlobalClassData(className);
  }
  public void removeGlobalClassData(String className) {
    GlobalClassData gcd = getGlobalClassData(className);
    if (gcd == null) {
      return;
    }
    buffer.removedGlobalClassName.add(className);
    getPersistenceService().delete(gcd);
  }
  public void storeGlobalClassData(GlobalClassData globalClassData) {
    buffer.removedGlobalClassName.remove(globalClassData.getClassName());
    getPersistenceService().save(globalClassData);
  }

  /*
   * DEPENDENT PACKAGES
   */
  public Set<PackageDependency> getPackageDependencies(String className) {
    return getPersistenceService().getPackageDependencies(className);
  }
  public void addPackageDependencies(PackageDefinitionUUID packageUUID, Set<String> classNames) {
    getPersistenceService().save(new PackageDependency(packageUUID, classNames));
  }
  public PackageDependency getPackageDependency(PackageDefinitionUUID packageId) {
    return getPersistenceService().getPackageDependency(packageId);
  }
  public void removePackageDependencies(PackageDefinitionUUID packageId) {
    PackageDependency packageDependency = getPackageDependency(packageId);
    if (packageDependency == null) {
      return;
    }
    getPersistenceService().delete(packageDependency);


  }

  public XpdlProcess findLatestProcessById(String subProcessId) {
    return getPersistenceService().findLatestProcessById(subProcessId);
  }

  public void addTimer(Timer timer) {
    EnvTool.getTimerSession().schedule(timer);
  }
  public List<Timer> getInstanceTimers(ProcessInstanceUUID instanceUUID) {
    return getPersistenceService().getInstanceTimers(instanceUUID);
  }
  public Collection<Timer> getExecutionTimers(XpdlExecution xpdlExecution) {
    return getPersistenceService().getExecutionTimers(xpdlExecution);
  }
  public void removeTimer(Timer timer) {
    EnvTool.getTimerSession().cancel(timer);
  }

  public synchronized long getNextActivityInstanceNb(ProcessInstanceUUID instanceUUID,
      String activityName) {
    ActivityInstanceKey nextPik = new ActivityInstanceKey(instanceUUID, activityName);
    getPersistenceService().save(nextPik);
    getPersistenceService().delete(nextPik);
    return nextPik.getNumber();
  }

  public synchronized long getNextProcessInstanceNb(ProcessDefinitionUUID processUUID) {
    ProcessInstanceKey nextPik = new ProcessInstanceKey(processUUID);
    getPersistenceService().save(nextPik);
    getPersistenceService().delete(nextPik);
    return nextPik.getNumber();
  }
}
