/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.persistence.xml;


import com.thoughtworks.xstream.XStream;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.PackageFullDefinitionImpl;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.runtime.impl.ProcessFullInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.services.Archivable;
import org.ow2.bonita.services.impl.AbstractHistory;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.novabpm.util.Misc;

/**
 *
 *
 * @author Pierre Vigneras
 * @date Feb 29, 2008
 */
public class XMLHistory extends AbstractHistory {

  private static final String XML_EXTENSION = ".xml";
  private static final String INSTANCE_FILENAME_PREFIX = "instance-";
  private static final String PACKAGE_FILENAME_PREFIX = "package-";

  private static final String DEFINITION_FOLDER = "definitions";
  private static final String RUNTIME_FOLDER = "runtime";
  private static final Logger LOG = Logger.getLogger(XMLHistory.class.getName());

  private void logWrite(final Archivable archivable, final File file) {
    if (LOG.isLoggable(Level.FINER)) {
      LOG.finer("Writing: " + archivable + " to file: " + file);
    }
    Writer writer = null;
    try {
      writer = new FileWriter(file);
      writer.append(xstream.toXML(archivable));
      writer.flush();
    } catch (IOException e) {
      throw new BonitaRuntimeException(e);
    } finally {
      Misc.close(writer);
    }
  }

  private Object logRead(final Class< ? extends Archivable> clazz, final File file) {
    if (LOG.isLoggable(Level.FINER)) {
      LOG.finer("Fetching class: " + clazz.getName() + " from file: " + file);
    }
    Reader reader = null;
    try {
      reader = new FileReader(file);
      return xstream.fromXML(reader);
    } catch (IOException e) {
      throw new BonitaRuntimeException(e);
    } finally {
      Misc.close(reader);
    }
  }

  private final File base;
  private final XStream xstream;

  public XMLHistory() {
    this(new File(Misc.TMP_DIR, "Bonita"));
  }

  public XMLHistory(final String path) {
    this(new File(path));
  }

  public XMLHistory(final File path) {
    this.base = new File(path, "history");
    LOG.config("Configuring history: " + XMLHistory.class.getName() + " with base: " + base);
    if (this.base.exists()) {
      if (!this.base.isDirectory() || !this.base.canRead() || !this.base.canWrite()) {
        throw new BonitaRuntimeException("Given history base: " + path + " should be a readable and writable directory!");
      }
    } else if (!this.base.mkdirs()) {
      throw new BonitaRuntimeException("Can't create base: " + path);
    }
    this.xstream = new XStream();
  }

  /*
   * PROCESS INSTANCE
   */

  public void archive(ProcessFullInstance processInst) {
    ProcessFullInstance newInstance = new ProcessFullInstanceImpl(processInst);
    //newInstance.end();
    final File file = getFile(processInst.getProcessDefinitionUUID(),
        processInst.getProcessInstanceUUID(), INSTANCE_FILENAME_PREFIX);
    checkDuplicate(newInstance, file);
    checkWritable(file);
    logWrite(newInstance, file);
  }
  
  public void remove(ProcessFullInstance processInst) {
    final File file = getFile(processInst.getProcessDefinitionUUID(),
        processInst.getProcessInstanceUUID(), INSTANCE_FILENAME_PREFIX);
    file.delete();
  }

  // points to: base/RUNTIME_FOLDER/processDefinitionUUID/INSTANCE_FILENAME_PREFIX+processUUID.xml
  private File getFile(final ProcessDefinitionUUID processUUID, final ProcessInstanceUUID instanceUUID, final String prefix) {
    Misc.checkArgsNotNull(processUUID, instanceUUID);
    final File runtimeFolderBase = new File(base, RUNTIME_FOLDER);
    File processFolder = new File(runtimeFolderBase, processUUID.toString());
    final File instanceFile = new File(processFolder, prefix + instanceUUID.toString() + XML_EXTENSION);
    return instanceFile;
  }

  public ProcessFullInstance getProcessInstance(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    final File runtimeFolder = new File(base, RUNTIME_FOLDER);
    final File[] processFolders = runtimeFolder.listFiles();
    if (processFolders == null) {
      return null;
    }
    for (File processFolder : processFolders) {
      File[] instanceFiles = processFolder.listFiles();
      if (instanceFiles != null) {
        for (File instanceFile : instanceFiles) {
          if (instanceFile.getName().equals(INSTANCE_FILENAME_PREFIX + instanceUUID.toString() + ".xml")) {
            return (ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile);
          }
        }
      }
    }
    return null;
  }

  public Set<ProcessFullInstance> getProcessInstances(ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    final Set<ProcessFullInstance> result = new HashSet<ProcessFullInstance>();
    if (processUUID != null) {
      final File runtimeFolder = new File(base, RUNTIME_FOLDER);
      File processFolder = new File(runtimeFolder, processUUID.toString());
      if (processFolder.exists()) {
        File[] instanceFiles = processFolder.listFiles();
        if (instanceFiles != null) {
          for (File instanceFile : instanceFiles) {
            result.add((ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile));
          }
        }
      }
    }
    return result;
  }

  /*public Set<ActivityFullInstance<ActivityBody>> getActivityInstances(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    ProcessFullInstance instance = getProcessInstance(instanceUUID);
    if (instance == null) {
      return new HashSet<ActivityFullInstance<ActivityBody>>();
    }
    Set<ActivityFullInstance<ActivityBody>> result = instance.getActivities();
    if (result == null) {
      return new HashSet<ActivityFullInstance<ActivityBody>>();
    }
    return result;
  }

  @SuppressWarnings("unchecked")
  public ActivityFullInstance<ActivityBody> getActivityInstance(ProcessInstanceUUID instanceUUID,
      String activityId, String iterationId) {
    Misc.checkArgsNotNull(instanceUUID, activityId, iterationId);
    ProcessFullInstance instance = getProcessInstance(instanceUUID);
    if (instance == null) {
      return null;
    }
    return instance.getActivity(activityId, iterationId);
  }
*/

  @SuppressWarnings("unchecked")
  public ActivityFullInstance< TaskInstance > getTaskInstance(TaskUUID taskUUID) {
    Misc.checkArgsNotNull(taskUUID);
    Set<ProcessFullInstance> instances = getAllInstances();
    if (instances != null) {
      for (ProcessFullInstance instance : instances) {
        Set<ActivityFullInstance<ActivityBody>> activities = getActivityInstances(instance.getProcessInstanceUUID());
        if (activities != null) {
          for (ActivityFullInstance< ? > activity : activities) {
            ActivityBody body = activity.getBody();
            if (body != null && body instanceof TaskInstance && body.getUUID().equals(taskUUID)) {
              return (ActivityFullInstance<TaskInstance>) activity;
            }
          }
        }
      }
    }
    return null;
  }

  private Set<ProcessFullInstance> getAllInstances() {
    Set<ProcessFullInstance> instances = new HashSet<ProcessFullInstance>();
    final File runtimeFolder = new File(base, RUNTIME_FOLDER);
    final File[] processFolders = runtimeFolder.listFiles();
    if (processFolders == null) {
      return null;
    }
    for (File processFolder : processFolders) {
      for (File instanceFile : processFolder.listFiles()) {
        instances.add((ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile));
      }
    }
    return instances;
  }

/*  @SuppressWarnings("unchecked")
  public Set<ActivityFullInstance< TaskInstance >> getTaskInstances(ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    final Set<ActivityFullInstance< TaskInstance >> result = new HashSet<ActivityFullInstance< TaskInstance >>();

    Set<ActivityFullInstance<ActivityBody>> activities = getActivityInstances(instanceUUID);
    if (activities != null) {
      for (ActivityFullInstance< ? > activity : activities) {
        ActivityBody body = activity.getBody();
        if (body != null && body instanceof TaskInstance) {
          result.add((ActivityFullInstance<TaskInstance>) activity);
        }
      }
    }
    return result;
  }
*/

//  private Object findAndRead(Class< ? extends Archivable> clazz, final String fileName) {
//    final File runtimeFolder = new File(base, RUNTIME_FOLDER);
//    final File[] perProcessFolders = runtimeFolder.listFiles();
//    if (perProcessFolders == null) {
//      return null;
//    }
//    for (File file : perProcessFolders) {
//      final File[] perDeploymentFolders = file.listFiles();
//      for (File instanceFolder : perDeploymentFolders) {
//        for (File candidateFolder : instanceFolder.listFiles()) {
//          for (File candidate : candidateFolder.listFiles()) {
//            final String name = candidate.getName();
//            if (name.equals(fileName)) {
//              return logRead(clazz, candidate);
//            }
//          }
//        }
//      }
//    }
//    return null;
//  }

  public File getBasePath() {
    return base;
  }


  private void checkWritable(final File file) {
    final File parent = file.getParentFile();
    if (!parent.exists() && !parent.mkdirs()) {
      // This test can fail because:
      // - mkdir failed because of security exception, or disk full ...
      // in this case an exception should be thrown
      // - the directory has been created after calling exists() test and before mkdir() has been called.
      // in this situation, there is no failure.
      // To solve this, we can add locks in this method,
      // or we can just check again that parent does not exist.
      // We choose the last solution.
      // TODO: see if the first call to exists() can be removed
      if (!parent.exists()) {
        throw new BonitaRuntimeException("Can't create path: " + parent);
      }
    }
  }

  private void checkDuplicate(final Archivable archivable, final File file) {
    if (file.exists()) {
      throw new IllegalArgumentException(" Can't archive: " + archivable
          + " --- A record with processDefinitionUUID: "
          + archivable + " has already been archived in file: " + file);
    }
  }

  public void clear() {
    if (LOG.isLoggable(Level.CONFIG)) {
      LOG.config("Clearing history: " + this + ". Deleting recursively directory: " + base);
    }
    Misc.deleteDir(getBasePath());
  }

  /*
   * DEFINITION
   */

  public void archive(PackageFullDefinition packageDef) {
    packageDef = new PackageFullDefinitionImpl(packageDef);
    //packageDef.end();
    final File file = getFile(packageDef.getUUID());
    checkWritable(file);
    logWrite(packageDef, file);
  }
  
  public void remove(PackageFullDefinition packageDef) {
    final File file = getFile(packageDef.getUUID());
    file.delete();
  }

//points to: base/DEFINITION_FOLDER/PACKAGE_FILENAME_PREFIX+packageUUID.xml
  private File getFile(PackageDefinitionUUID packageUUID) {
    final File definitionFolder = new File(base, DEFINITION_FOLDER);
    final File packageFile = new File(definitionFolder, PACKAGE_FILENAME_PREFIX + packageUUID + XML_EXTENSION);
    return packageFile;
  }

  public Set<PackageFullDefinition> getPackages() {
    final Set<PackageFullDefinition> result = new HashSet<PackageFullDefinition>();
    final File definitionFolder = new File(base, DEFINITION_FOLDER);
    final File[] packageFiles = definitionFolder.listFiles();
    if (packageFiles == null) {
      return result;
    }
    for (File packageFile : packageFiles) {
      result.add((PackageFullDefinition) logRead(PackageFullDefinition.class, packageFile));
    }
    return result;
  }

/*  public Set<PackageFullDefinition> getPackages(String packageId) {
    Misc.checkArgsNotNull(packageId);
    Set<PackageFullDefinition> packages = getPackages();
    Set<PackageFullDefinition> result = new HashSet<PackageFullDefinition>();
    if (packages != null) {
      for (PackageFullDefinition package1 : packages) {
        if (package1.getPackageId().equals(packageId)) {
          result.add(package1);
        }
      }
    }
    return result;
  }*/

  public PackageFullDefinition getPackage(PackageDefinitionUUID packageUUID) {

    Misc.checkArgsNotNull(packageUUID);
    File file = getFile(packageUUID);
    if (!file.exists()) {
      return null;
    }
    return (PackageFullDefinition) logRead(PackageFullDefinition.class, file);
  }

/*  public ProcessFullDefinition getPackageProcess(
      PackageDefinitionUUID packageDefinitionUUID, String processId) {

    Misc.checkArgsNotNull(packageDefinitionUUID, processId);
    PackageFullDefinition packageDef = getPackage(packageDefinitionUUID);
    if (packageDef != null) {
      Set<ProcessFullDefinition> processes = packageDef.getProcesses();
      for (ProcessFullDefinition process : processes) {
        if (process.getProcessId().equals(processId)) {
          return process;
        }
      }
    }
    return null;
  }
*/

  public ActivityFullInstance<ActivityBody> getActivityInstance(
      ActivityInstanceUUID activityInstanceUUID) {

    Misc.checkArgsNotNull(activityInstanceUUID);
    Set<ProcessFullInstance> instances = getAllInstances();
    if (instances != null) {
      for (ProcessFullInstance instance : instances) {
        Set<ActivityFullInstance<ActivityBody>> activities = instance.getActivities();
        if (activities != null) {
          for (ActivityFullInstance<ActivityBody> activity : activities) {
            if (activity.getUUID().equals(activityInstanceUUID)) {
              return activity;
            }
          }
        }
      }
    }
    return null;
  }

  public Set<ActivityFullInstance<TaskInstance>> getUserInstanceTasks(
      String userId, ProcessInstanceUUID instanceUUID, TaskState taskState) {
    Misc.checkArgsNotNull(instanceUUID, taskState);
    ProcessFullInstance instance = getProcessInstance(instanceUUID);
    Set<ActivityFullInstance<TaskInstance>> result = new HashSet<ActivityFullInstance<TaskInstance>>();
    
    if (instance == null) {
      return result;
    }
    
    result.addAll(getUserTasks(instance.getActivities(), userId, taskState));

    return result;
  }

  @SuppressWarnings("unchecked")
  private Set<ActivityFullInstance<TaskInstance>> getUserTasks(Set<ActivityFullInstance<ActivityBody>> activities,
      String userId, TaskState taskState) {
    Set<ActivityFullInstance<TaskInstance>> result = new HashSet<ActivityFullInstance<TaskInstance>>();
    if (activities != null) {
      for (ActivityFullInstance< ? > activity : activities) {
        ActivityBody body = activity.getBody();
        if (body != null && body instanceof TaskInstance) {
          TaskInstance taskInstance = (TaskInstance)body;
          boolean userIdOk = false;
          try {
            String taskUserId = taskInstance.getTaskUser();
            userIdOk = taskUserId != null && taskUserId.equals(userId);
          } catch (IllegalStateException e) {
            // nothing
          }
          boolean candidatesOk = false;
          try {
            Set<String> taskCandidates = taskInstance.getTaskCandidates();
            candidatesOk = taskCandidates != null && taskCandidates.contains(userId);
          } catch (IllegalStateException e) {
            // nothing
          }
          boolean stateOk = taskInstance.getState().equals(taskState);
          if ((userIdOk || candidatesOk) && stateOk) {
            result.add((ActivityFullInstance<TaskInstance>) activity);
          }
        }
      }
    }
    return result;
  }

  public Set<ActivityFullInstance<TaskInstance>> getUserTasks(String userId, TaskState taskState) {
    Set<ActivityFullInstance<TaskInstance>> result = new HashSet<ActivityFullInstance<TaskInstance>>();
    Set<ProcessFullInstance> instances = getAllInstances();
    if (instances != null) {
      for (ProcessFullInstance instance : instances)   {
        Set<ActivityFullInstance<ActivityBody>> activities = instance.getActivities();
        if (activities != null) {
          Set<ActivityFullInstance<TaskInstance>> taskActivities = getUserTasks(activities, userId, taskState);
          if (taskActivities != null) {
            result.addAll(taskActivities);
          }
        }
      }
    }
    return result;
  }
}
