/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.ClassInfo;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.definition.activity.HookExecutor;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition.ParticipantType;
import org.ow2.bonita.facade.runtime.TaskState;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.StandaloneAccessorUtil;

/**
 * is one task instance that can be assigned to an user
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class TaskRunTime {

  private static final long serialVersionUID = 1L;
  protected long dbid;

  private static Logger log = Logger.getLogger(TaskRunTime.class.getName());

  protected TaskUUID taskUUID;
  protected String description = null;
  protected String userId = null;
  protected Date create = null;
  protected Date start = null;
  protected Date end = null;
  protected Date dueDate = null;

  protected TaskState state = TaskState.INITIAL;
  protected TaskState stateBeforeSuspend = null;

  protected AbstractActivity activityDef;
  protected String iterationId;
  protected String activityInstanceId;
  protected String activityId;
  protected PackageDefinitionUUID packageUUID;
  protected ProcessInstanceUUID instanceUUID;
  protected XpdlExecution xpdlExecution;

  // List of users having this role
  protected Set<String> candidates = null;


  protected TaskRunTime() {
  }

  public TaskRunTime(TaskUUID taskUUID, String nodeName,
      AbstractActivity abstractActivity, XpdlExecution xpdlExecution) {
    this.taskUUID = taskUUID;
    this.activityDef = abstractActivity;
    this.activityId = nodeName;
    this.xpdlExecution = xpdlExecution;
    this.instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
    this.packageUUID = xpdlExecution.getProcessDefinition().getPackageDefinitionUUID();
    this.iterationId = xpdlExecution.getIterationId();
    this.activityInstanceId = xpdlExecution.getActivityInstanceId();
  }

  public void resolve(Performer performer) {
    mapRole(performer);
    assign(performer);
  }

  protected void assign(Performer performer) {
    ClassInfo performerAssignInfo = performer.getPerformerAssign();
    if (performerAssignInfo != null) {
      PerformerAssign performerAssign = ClassDataLoader.getPerformerAssignInstance(packageUUID, performerAssignInfo);
      try {
        this.userId = performerAssign.selectUser(StandaloneAccessorUtil.getQueryAPIAccessor(),
            instanceUUID, activityId, iterationId, activityInstanceId, this.candidates);
      } catch (Exception e) {
        throw new BonitaRuntimeException("Exception caught while executing performerAssign : " + performerAssign, e);
      }
    }
  }

  protected void mapRole(Performer performer) {
    ClassInfo rolemapperInfo = performer.getRoleMapper();
    if (rolemapperInfo != null) {
      RoleMapper roleMapper = ClassDataLoader.getRoleMapperInstance(packageUUID, rolemapperInfo);
      try {
        this.candidates = roleMapper.searchMembers(
            StandaloneAccessorUtil.getQueryAPIAccessor(), instanceUUID, performer.getParticipantId());
      } catch (Exception e) {
        throw new BonitaRuntimeException("Exception caught while executing role mapper : " + roleMapper, e);
      }
    }
  }

  public void ready() {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Ready task : " + taskUUID + " on activity " + activityId);
    }
    this.state = TaskState.READY;

    // get performer (aka participant/swimlane)
    Performer performer = activityDef.getPerformer();

    Set<String> ar = new HashSet<String>();
    //if human set userId and candidates
    if (performer.getParticipantType().equals(ParticipantType.HUMAN)) {
      this.userId = performer.getParticipantId();
      ar.add(performer.getParticipantId());
      this.candidates = ar;
    }
    //Only ROLE and ORGANIZATIONAL_UNIT can have Role mapper (ParticipantBinding)
    if (performer.getRoleMapper() != null) {
      mapRole(performer);
    } else {
      // no role mapper: set an empty candidates list
      this.candidates = ar;
    }

    // performer assignment associated to this task
    assign(performer);

    EnvTool.getRecorder().recordTaskReady(this);

    HookExecutor.executeHooks(activityDef.getJavaHooks(), xpdlExecution, activityId, JavaHook.Type.taskOnReady);
  }

  public void start(String userId, boolean assignTask) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Task : " + taskUUID + " on activity " + activityId + " is ready");
    }
    final Recorder recorder = EnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(this, userId, candidates, userId);
    }
    this.state = TaskState.EXECUTING;
    recorder.recordTaskStarted(this, userId);
    recorder.recordBodyStarted(activityDef, xpdlExecution);
    HookExecutor.executeHooks(activityDef.getJavaHooks(), xpdlExecution, activityId, JavaHook.Type.taskOnStart);
  }

  public void finish(String userId, boolean assignTask) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Finishing task : " + taskUUID + " on activity " + activityId);
    }
    final Recorder recorder = EnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(this, userId, candidates, userId);
    }
    this.state = TaskState.FINISHED;
    recorder.recordTaskFinished(this, EnvTool.getUserId());
    HookExecutor.executeHooks(activityDef.getJavaHooks(), xpdlExecution, activityId, JavaHook.Type.taskOnFinish);
    activityDef.signal(xpdlExecution, AbstractActivity.BODY_FINISHED, null);
    this.xpdlExecution = null;
  }

  public void suspend(String userId, boolean assignTask) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Suspending task : " + taskUUID + " on activity " + activityId);
    }
    final Recorder recorder = EnvTool.getRecorder();
    if (assignTask) {
      this.userId = userId;
      recorder.recordTaskAssigned(this, userId, candidates, userId);
    }
    this.stateBeforeSuspend = this.state;
    this.state = TaskState.SUSPENDED;
    recorder.recordTaskSuspended(this, userId);
    HookExecutor.executeHooks(activityDef.getJavaHooks(), xpdlExecution, activityId, JavaHook.Type.taskOnSuspend);
  }

  public void resume(String resumeUserId, boolean taskAssign) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Resuming task : " + taskUUID + " on activity " + activityId);
    }
    final Recorder recorder = EnvTool.getRecorder();
    if (taskAssign) {
      this.userId = resumeUserId;
      recorder.recordTaskAssigned(this, resumeUserId, candidates, userId);
    }
    this.state = this.stateBeforeSuspend;
    recorder.recordTaskResumed(this, resumeUserId, this.userId);
    HookExecutor.executeHooks(activityDef.getJavaHooks(), xpdlExecution, activityId, JavaHook.Type.taskOnResume);
  }

  public void assign(java.util.Set<java.lang.String> candidates, String assignUserId) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("assigning task : " + taskUUID + " on activity " + activityId);
    }
    final Recorder recorder = EnvTool.getRecorder();
    this.userId = assignUserId;
    this.candidates = candidates;
    recorder.recordTaskAssigned(this, EnvTool.getUserId(), candidates, assignUserId);
  }

  public String getDescription() {
    return description;
  }

  public void setDescription(String description) {
    this.description = description;
  }

  public String getUserId() {
    return userId;
  }

  public Date getCreate() {
    return create;
  }

  public void setCreate(Date create) {
    this.create = create;
  }

  public Date getDueDate() {
    return dueDate;
  }

  public void setDueDate(Date dueDate) {
    this.dueDate = dueDate;
  }

  public Date getEnd() {
    return end;
  }

  public void setEnd(Date end) {
    this.end = end;
  }
  public Date getStart() {
    return start;
  }
  public void setStart(Date start) {
    this.start = start;
  }
  public TaskUUID getUUID() {
    return taskUUID;
  }
  public Set<String> getCandidates() {
    return candidates;
  }
  public void setCandidates(Set<String> candidates) {
    this.candidates = candidates;
  }
  public TaskState getState() {
    return state;
  }
  public void setState(TaskState state) {
    this.state = state;
  }

  @Override
  public String toString() {
    return "TaskRuntime : userId = " + userId + ", candidates = " + candidates
      + ", activityId = " + activityId;
  }


  public String getActivityId() {
    return activityId;
  }

  public ProcessInstanceUUID getInstanceUUID() {
    return instanceUUID;
  }

  public long getDbid() {
    return dbid;
  }


  public XpdlExecution getXpdlExecution() {
    return xpdlExecution;
  }


  public AbstractActivity getActivityDef() {
    return activityDef;
  }


  public PackageDefinitionUUID getPackageUUID() {
    return packageUUID;
  }

}
