/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.services.impl;

import java.util.Date;
import org.jbpm.pvm.internal.model.NodeImpl;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.TaskFullInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.VariableUpdate;
import org.ow2.bonita.facade.runtime.impl.ActivityBodyImpl;
import org.ow2.bonita.facade.runtime.impl.ActivityFullInstanceImpl;
import org.ow2.bonita.facade.runtime.impl.ProcessFullInstanceImpl;
import org.ow2.bonita.facade.runtime.impl.VariableUpdateFullImpl;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Journal;
import org.ow2.novabpm.util.Misc;

/**
 * @author Pierre Vigneras, Guillaume Porcher
 *
 */
public abstract class AbstractJournal implements Journal {

  public abstract void storeInstanceRecord(ProcessFullInstanceImpl instanceRecord);
  public abstract void storePackageDef(PackageFullDefinition packageDef);

  public void recordEnterActivity(final AbstractActivity activity, final XpdlExecution xpdlExecution) {
    Misc.checkArgsNotNull(xpdlExecution);
    final NodeImpl node = xpdlExecution.getNode();
    XpdlInstance instance = xpdlExecution.getXpdlInstance();
    final String id = node.getName();
    final PackageDefinitionUUID packageUUID = instance.getPackageDefinitionUUID();
    final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
    final ProcessInstanceUUID instanceUUID = instance.getUUID();
    final String iterationId = xpdlExecution.getIterationId();
    final String activityInstanceId = xpdlExecution.getActivityInstanceId();


    final ActivityFullInstanceImpl<ActivityBody> activityRecord = new ActivityFullInstanceImpl<ActivityBody>(
        new ActivityInstanceUUID(instanceUUID, activity.getActivityId()), id,
        packageUUID, processUUID, instanceUUID, iterationId, activityInstanceId);

    ActivityBody body = activity.getBody(xpdlExecution, activityRecord);

    activityRecord.setBody(body);
    ProcessFullInstance processInstance = getProcessInstance(instanceUUID);
    processInstance.addActivity(activityRecord);
    activityRecord.setVariablesBeforeReady(xpdlExecution.getScopeVariables());
  }

  public void recordBodyStarted(AbstractActivity activity, XpdlExecution xpdlExecution) {
    final String activityId = activity.getActivityId();
    ProcessInstanceUUID instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
    final String iterationId = xpdlExecution.getIterationId();
    final String activityInstanceId = xpdlExecution.getActivityInstanceId();
    ActivityFullInstance<ActivityBody> activityInstance =
      getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId);
    ((ActivityBodyImpl)activityInstance.getBody()).setStartedDate(new Date());
  }

  public void recordBodyEnded(AbstractActivity activity, XpdlExecution xpdlExecution) {
    final String activityId = activity.getActivityId();
    ProcessInstanceUUID instanceUUID = xpdlExecution.getXpdlInstance().getUUID();
    final String iterationId = xpdlExecution.getIterationId();
    final String activityInstanceId = xpdlExecution.getActivityInstanceId();
    ActivityFullInstance<ActivityBody> activityInstance =
      getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId);
    ((ActivityBodyImpl)activityInstance.getBody()).setEndedDate(new Date());
  }

  public void recordInstanceEnded(XpdlInstance xpdlInstance, final String userId) {
    Misc.checkArgsNotNull(xpdlInstance);
    final ProcessFullInstanceImpl instanceRecord =
      (ProcessFullInstanceImpl) getProcessInstance(xpdlInstance.getUUID());
    Misc.badStateIfNull(instanceRecord, "Ouch! Can't find the ProcessInstance for: " + xpdlInstance);
    instanceRecord.setEndedDate(new Date());
    instanceRecord.setEndedBy(userId);
  }

  public void recordInstanceStarted(XpdlInstance instance, final String userId) {
    Misc.checkArgsNotNull(instance);
    final ProcessFullInstanceImpl instanceRecord =
      new ProcessFullInstanceImpl(instance.getUUID(),
          instance.getPackageDefinitionUUID(), instance.getProcessDefinitionUUID(),
          (instance.getParentInstance() != null) ? instance.getParentInstance().getUUID() : null,
              instance.getRootExecution().getVariables());
    instanceRecord.setStartedDate(new Date());
    Misc.badStateIfNull(userId, "Ouch! Can't get the user id from the environment!");
    instanceRecord.setStartedBy(userId);
    storeInstanceRecord(instanceRecord);
  }

  public void recordTaskFinished(TaskRunTime task, final String userId) {
    Misc.checkArgsNotNull(task);
    TaskUUID taskUUID = task.getUUID();
    ActivityFullInstance< TaskInstance > activity = getTaskInstance(taskUUID);
    TaskFullInstance taskInst = (TaskFullInstance) activity.getBody();
    Misc.badStateIfNull(taskInst, "Ouch! Can't find the TaskInstance for: " + taskInst.getUUID());
    taskInst.setEndedBy(userId);
    taskInst.setTaskState(task.getState(), userId);
  }

  public void recordTaskReady(TaskRunTime task) {
    Misc.checkArgsNotNull(task);
    TaskUUID taskUUID = task.getUUID();
    ActivityFullInstance< TaskInstance > activity = getTaskInstance(taskUUID);
    TaskFullInstance taskRecord = (TaskFullInstance) activity.getBody();
    Misc.badStateIfNull(taskRecord, "Ouch! Can't find the TaskInstance for: " + task.getUUID());
    taskRecord.setTaskState(task.getState(), "SYSTEM");
    taskRecord.setInitialCandidates(task.getCandidates());
    taskRecord.setInitialUserId(task.getUserId());
  }

  public void recordTaskStarted(TaskRunTime task, final String userId) {
    Misc.checkArgsNotNull(task);
    TaskUUID taskUUID = task.getUUID();
    ActivityFullInstance< TaskInstance > activity = getTaskInstance(taskUUID);
    TaskFullInstance taskRecord = (TaskFullInstance) activity.getBody();
    Misc.badStateIfNull(taskRecord, "Ouch! Can't find the TaskInstance for: " + task.getUUID());
    taskRecord.setStartedBy(userId);
    taskRecord.setTaskState(task.getState(), userId);
    //taskRecord.setUserId(userId);
  }

  public void recordTaskResumed(TaskRunTime taskRunTime, String userId, String userIdBeforeSuspend) {
    Misc.checkArgsNotNull(taskRunTime);
    TaskFullInstance taskInst = getTask(taskRunTime);
    taskInst.setTaskState(taskRunTime.getState(), userId);
    //taskInst.setUserId(userIdBeforeSuspend);
  }
  public void recordTaskSuspended(final TaskRunTime taskRunTime, final String userId) {
    Misc.checkArgsNotNull(taskRunTime);
    TaskFullInstance taskInst = getTask(taskRunTime);
    taskInst.setTaskState(taskRunTime.getState(), userId);
    //taskInst.setUserId(userId);
  }

  public void recordTaskAssigned(TaskRunTime taskRunTime, String assignerUserId,
      java.util.Set<java.lang.String> candidates, String userId) {
    Misc.checkArgsNotNull(taskRunTime);
    TaskFullInstance taskFullInst = getTask(taskRunTime);
    taskFullInst.setTaskAssign(taskRunTime.getState(), assignerUserId, candidates, userId);
  }

  private TaskFullInstance getTask(TaskRunTime taskRuntime) {
    TaskUUID taskUUID = taskRuntime.getUUID();
    ActivityFullInstance< TaskInstance > taskactivity = getTaskInstance(taskUUID);
    Misc.badStateIfNull(taskactivity, "Ouch! Can't find the taskActivity with taskUUID for: " + taskRuntime.getUUID());
    TaskFullInstance taskInst = (TaskFullInstance)taskactivity.getBody();
    return taskInst;
  }
  /*
  private String getLastUserId(TaskFullInstance taskInst) {
    List<StateUpdate> stateUpdates = taskInst.getStateUpdates();
    int size = stateUpdates.size();
    for (int i = size - 1; i >= 0; i--) {
      StateUpdate stateUpdate = stateUpdates.get(i);
      TaskState finalState = stateUpdate.getFinalState();
      if (finalState.equals(TaskState.READY) || finalState.equals(TaskState.EXECUTING)) {
        return stateUpdate.getUserId();
      }
    }
    return null;
  }
  */
  public void recordPackageDeployed(PackageFullDefinition packageDef) {
    storePackageDef(packageDef);
  }

  public void recordPackageUndeployed(PackageFullDefinition packageDef, String userId) {
    packageDef.setUndeployedBy(userId);
    packageDef.setUndeployedDate(new Date());
  }

  public void recordActivityVariableUpdated(String variableId, Object variableValue, ProcessInstanceUUID instanceUUID,
      String activityId, String iterationId, String activityInstanceId, String userId) {
    ActivityFullInstance<ActivityBody> activityInst =
      getActivityInstance(instanceUUID, activityId, iterationId, activityInstanceId);
    VariableUpdate varUpdate = new VariableUpdateFullImpl(new Date(), userId, variableId, variableValue);
    activityInst.addVariableUpdate(varUpdate);
  }

  public void recordInstanceVariableUpdated(String variableId, Object variableValue,
      ProcessInstanceUUID instanceUUID, String userId) {
    ProcessFullInstance processInst = getProcessInstance(instanceUUID);
    VariableUpdate varUpdate = new VariableUpdateFullImpl(new Date(), userId, variableId, variableValue);
    processInst.addVariableUpdate(varUpdate);
  }
}
