/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.ow2.bonita.env.EnvironmentNotFoundException;
import org.ow2.bonita.facade.APIAccessor;
import org.ow2.bonita.facade.CommandAPI;
import org.ow2.bonita.facade.Context;
import org.ow2.bonita.facade.DefinitionAPI;
import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.QueryDefinitionAPI;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.ejb.EJBCommandAPIInterceptor;
import org.ow2.bonita.facade.ejb.ejb2.internal.EJB2AccessorUtil;
import org.ow2.bonita.facade.impl.APIAccessorImpl;
import org.ow2.novabpm.util.Misc;

/**
 *This utility class has been provided to unify access to Bonita APIs and to avoid the use of lookups
 *in JEE deployments: org.ow2.bonita.util.AccessorUtil.<br>
 *Through this class, Nova Bonita APIs can be reached in a unified way in both local and remote applications.<br>
 *For that to be done, the system property called "org.ow2.bonita.api-type" must be defined at
 *client side to specify whether the APIs will be reached locally or remotely<br>
 *(possible values are "standard","auto-detect",''ejb2" and "ejb3").<br>
 */

public final class AccessorUtil {

  public static final String API_TYPE_PROPERTY = "org.ow2.bonita.api-type";
  public static final String RUNTIMEAPI_JNDINAME = "runtimeAPI";
  public static final String MANAGEMENT_JNDINAME = "managementAPI";
  public static final String DEFINITIONAPI_JNDINAME = "definitionAPI";
  public static final String QUERYRUNTIME_JNDINAME = "queryRuntimeAPI";
  public static final String QUERYDEFINITION_JNDINAME = "queryDefinitionAPI";
  public static final String COMMANDAPI_JNDINAME = "commandAPI";

  private static final Logger LOG = Logger.getLogger(AccessorUtil.class.getName());
  private static Context context;
  static final APIAccessor STANDALONE_ACCESSOR = new APIAccessorImpl();

  private static InitialContext initialContext;
  static {
    //First initialize the context, next check how to access api
    try {
      initialContext = new InitialContext();
    } catch (NamingException ne) {
      throw new ExceptionInInitializerError(ne);
    }

    context = findContext();
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("Using: " + context + " as the api-type.");
    }
  }

  static Context findContext() {
    String apiType = System.getProperty(API_TYPE_PROPERTY);
    if (apiType == null) {
      if (LOG.isLoggable(Level.WARNING)) {
        LOG.warning("Property: " + API_TYPE_PROPERTY + " has not been specified for api-type."
            + " Possible values are: " + Arrays.asList(Context.values())
            + ". Using default value: " + Context.AutoDetect);
      }
      apiType = Context.AutoDetect.toString();
    }
    final Context result = Misc.stringToEnum(Context.class, apiType);
    if (!result.equals(Context.AutoDetect)) {
      if (LOG.isLoggable(Level.CONFIG)) {
        LOG.config("API-Type: " + result + " has been specified through property: " + API_TYPE_PROPERTY);
      }
      return result;
    } 
    return autoDetectContext();
  }


  /* Search in which mode we should access to API :
   * First search in the environment.
   * If not present, try to lookup EJB3 RuntimeAPI interface and after EJB2 RuntimeAPI interface
   * if EJB3 interface not available.
   * Assume that we are in an J2SE mode if none of these are available.
   */
  private static Context autoDetectContext() {
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("Auto detecting api-type.");
    }
    try {
      if (LOG.isLoggable(Level.INFO)) {
        LOG.info("Looking into environment the specified api-type.");
      }
      final Context result = Context.getFromEnvironment();
      if (LOG.isLoggable(Level.INFO)) {
        LOG.info("API-Type: " + result + " has been taken from the environment." 
            + " (Note that the api-type can be set to: " + Context.AutoDetect 
            + " in the environment. So this api-type may be the output of the autodetecting algorithm)."
            + " Increase " + Context.class.getName() + " logging level to " + Level.FINER 
            + " for details.");
      }
      return result;
    } catch (EnvironmentNotFoundException iee) {
      return Context.autoDetect();
    }
  }

  private AccessorUtil() { }

  public static APIAccessor getAPIAccessor() {
    return STANDALONE_ACCESSOR;
  }
  
  @SuppressWarnings("unchecked")
  public static <T> T lookup(String name) {
    try {
      return (T) initialContext.lookup(name);
    } catch (NamingException e) {
      throw new BonitaRuntimeException(e);
    }
  }

  /**
   * To get the RuntimeAPI interface with a unify client access.
   * @return the RuntimeAPI interface.
   */
  public static RuntimeAPI getRuntimeAPI() {
    switch (context) {
      case EJB2 :
        return EJB2AccessorUtil.getRuntimeAPI();
      case EJB3 :
        return lookup(RUNTIMEAPI_JNDINAME);
      default :
        return getAPIAccessor().getRuntimeAPI();
    }
  }

  /**
   * To get the ManagementAPI interface with a unify client access.
   * @return the ManagementAPI interface.
   */
  public static ManagementAPI getManagementAPI() {
    switch (context) {
      case EJB2 :
        return EJB2AccessorUtil.getManagementAPI();
      case EJB3 :
        return lookup(MANAGEMENT_JNDINAME);
      default :
        return getAPIAccessor().getManagementAPI();
    }
  }

  /**
   * To get the DefinitionAPI interface with a unify client access.
   * @return the DefinitionAPI interface.
   */
  public static DefinitionAPI getDefinitionAPI() {
    switch (context) {
      case EJB2 :
        return EJB2AccessorUtil.getDefinitionAPI();
      case EJB3 :
        return lookup(DEFINITIONAPI_JNDINAME);
      default :
        return getAPIAccessor().getDefinitionAPI();
    }
  }

  /**
   * To get the QueryRuntimeAPI interface with a unify client access.
   * @return the QueryRuntimeAPI interface.
   */
  public static QueryRuntimeAPI getQueryRuntimeAPI() {
    switch (context) {
      case EJB2 :
        return EJB2AccessorUtil.getQueryRuntimeAPI();
      case EJB3 :
        return lookup(QUERYRUNTIME_JNDINAME);
      default :
        return getAPIAccessor().getQueryRuntimeAPI();
    }
  }

  /**
   * To get the QueryDefinitionAPI interface with a unify client access.
   * @return the QueryDefinitionAPI interface.
   */
  public static QueryDefinitionAPI getQueryDefinitionAPI() {
    switch (context) {
      case EJB2 :
        return EJB2AccessorUtil.getQueryDefinitionAPI();
      case EJB3 :
        return lookup(QUERYDEFINITION_JNDINAME);
      default :
        return getAPIAccessor().getQueryDefinitionAPI();
    }
  }

  /**
   * To get the CommandAPI interface with a unify client access.
   * @return the CommandAPI interface.
   */
  public static CommandAPI getCommandAPI() {
    CommandAPI ejbCommandAPI = null;
    switch (context) {
      case EJB2 :
        ejbCommandAPI = EJB2AccessorUtil.getCommandAPI();
        break;
      case EJB3 :
        ejbCommandAPI = lookup(COMMANDAPI_JNDINAME);
        break;
      default :
        return getAPIAccessor().getCommandAPI();
    } 
    return new EJBCommandAPIInterceptor(ejbCommandAPI);
  }
}


