/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.session.TimerSession;
import org.ow2.bonita.persistence.XpdlPersistenceService;
import org.ow2.bonita.runtime.Repository;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.impl.QuerierListAccessor;
import org.ow2.novabpm.util.Misc;

/**
 * This class holds the {@link EnvironmentFactory} singleton.
 *
 * The {@link EnvironmentFactory} returned by {@link #getEnvironmentFactory()} comes from the
 * parsing of the resource defined by the property
 * {@link #ENVIRONMENT_PROPERTY}. If this property has not been set, the
 * {@link #DEFAULT_ENVIRONMENT} resource is used and parsed.
 *
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public final class EnvTool {

  private EnvTool() { }

  private static Environment getEnv() {
    final Environment environment = Environment.getCurrent();
    Misc.badStateIfNull(environment, "Environment is null!");
    return environment;
  }
  private static <T> T getEnvClass(final Class<T> clazz) {
    final Environment environment = getEnv();
    final T object = environment.get(clazz);
    if (object == null) {
      throw new BonitaRuntimeException("No object of class " + clazz.getName() + " found in environment!");
    }
    return object;
  }

  @SuppressWarnings("unchecked")
  private static <T> T getEnvObject(final Class<T> clazz, final String name) {
    final Environment environment = getEnv();

    Misc.badStateIfNull(environment, "Environment is null!");
    final Object object = environment.get(name);
    if (object == null) {
      throw new BonitaRuntimeException("No object named " + name + " found in environment!");
    } else if (!clazz.isInstance(object)) {
      throw new BonitaRuntimeException("Object: " + object + " named: " + name
          + " found in environment is not an instance of class " + clazz.getName() + "!");
    }
    return (T) object;
  }

  public static Repository getRepository() {
    return getEnvClass(Repository.class);
  }

  public static Recorder getRecorder() {
    return getEnvObject(Recorder.class, Recorder.DEFAULT_KEY);
  }
  public static Archiver getArchiver() {
    return getEnvObject(Archiver.class, Archiver.DEFAULT_KEY);
  }

  private static QuerierListAccessor getQuerierListAccessor() {
    return EnvTool.get(QuerierListAccessor.class, Querier.DEFAULT_KEY);
  }
  public static Querier getAllQueriers() {
    return getQuerierListAccessor().getAllQueriers();
  }
  public static Querier getJournalQueriers() {
    return getQuerierListAccessor().getJournals();
  }
  public static Querier getHistoryQueriers() {
    return getQuerierListAccessor().getHistories();
  }
  public static XpdlPersistenceService getPersistenceService(String id) {
    return getEnvObject(XpdlPersistenceService.class, id);
  }

  public static String getUserId() {
    return getEnv().getUserId();
  }
/*  public static void setUserId(final String userId) {
    getEnv().setUserId(userId);
  }*/

  @SuppressWarnings("unchecked")
  public static <T> T get(final Class<T> class1, final String key) {
    Misc.checkArgsNotNull(class1, key);
    final Object o = getEnv().get(key);
    Misc.badStateIfNull(o, "Object '" + key + "' can't be found in environment");
    Misc.badStateIfFalse(class1.isAssignableFrom(o.getClass()),
        "Object got from environment is not an instance of " + class1
        + ". It is  : " + o.getClass());
    return (T) o;
  }

  public static TimerSession getTimerSession() {
    return getEnvClass(TimerSession.class);
  }

}
