/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jbpm.pvm.PvmException;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.type.Converter;
import org.jbpm.pvm.internal.type.Type;
import org.jbpm.pvm.internal.type.Variable;
import org.jbpm.pvm.internal.type.VariableTypeResolver;
import org.jbpm.pvm.internal.type.variable.NullVariable;
import org.jbpm.pvm.internal.type.variable.UnpersistableVariable;
import org.ow2.bonita.facade.def.majorElement.PackageFullDefinition;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.runtime.var.Enumeration;
import org.ow2.bonita.services.Archivable;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Recorder;
import org.ow2.novabpm.util.Misc;
import org.ow2.util.asm.AnnotationVisitor;
import org.ow2.util.asm.Attribute;
import org.ow2.util.asm.ClassReader;
import org.ow2.util.asm.ClassVisitor;
import org.ow2.util.asm.FieldVisitor;
import org.ow2.util.asm.MethodVisitor;

/**
 * @author Pierre Vigneras
 * @date Mar 6, 2008
 */
public abstract class Tool {

  public static final String DEPLOYMENT_IDPREFIX = "deployment$";

  private Tool() { }

  private static Object deploymentMutex = new Object();

  public static Object getDeploymentMutex() {
    return deploymentMutex;
  }

  /**
   * @author Pierre Vigneras
   * @date Mar 6, 2008
   */
  public static class AtomicArchiveException extends BonitaRuntimeException {

    public AtomicArchiveException(final String msg, final Throwable cause) {
      super(msg, cause);
    }
  }

  public static class MyVisitor implements ClassVisitor {
    private String className = null;
    private String superClassName = null;
    private String[] interfaces = null;
    
    public void visit(int version, int access, String name, String signature,
        String superName, String[] interfaces) {
      this.className = name;
      this.interfaces = interfaces;
      this.superClassName = superName;
    }
    public void visitSource(String source, String debug) { }
    public void visitOuterClass(String owner, String name, String desc) { }
    public AnnotationVisitor visitAnnotation(String desc, boolean visible) { return null; }
    public void visitAttribute(Attribute attr) { }
    public void visitInnerClass(String name, String outerName, String innerName,
        int access) { }
    public FieldVisitor visitField(int access, String name, String desc,
        String signature, Object value) { return null; }
    public MethodVisitor visitMethod(int access, String name, String desc,
        String signature, String[] exceptions) { return null; }
    public void visitEnd() { }

    
    public String getClassName() {
      return this.className.replace("/", ".");
    }
    public String getSuperClassName() {
      return superClassName;
    }
    public String[] getInterfaces() {
      return interfaces;
    }
  }

  /**
   * Perform a all-or-nothing archive.
   *
   * This method does the following: it first tries to archive the given Archivable
   * into the specified {@link Archiver}. If it succeeds and only in this case, it tries
   * to remove the given record from the specified 
   *
   * @param record a record
   * @throws AtomicArchiveException if archiving failed.
   */
  public static void atomicArchive(final Archivable archivable,
      final Recorder recorder,
      final Archiver archiver) {
    if (!(archivable instanceof PackageFullDefinition) && !(archivable instanceof ProcessFullInstance)) {
      throw new IllegalArgumentException("Don't know how to archivable: "
          + archivable + (archivable != null ? " of type: " + archivable.getClass().getName() : ""));
    }
    try {
      if (archivable instanceof PackageFullDefinition) {
        archiver.archive((PackageFullDefinition) archivable);
      } else {
        // archivable instanceof ProcessFullInstance because of previous checks
        archiver.archive((ProcessFullInstance) archivable);
      }
    } catch (Exception e) {
      throw new AtomicArchiveException("Can't archive instance archivable: "
          + archivable + " into archiver: " + archiver, e);
    }
    try {
      if (archivable instanceof PackageFullDefinition) {
        recorder.remove((PackageFullDefinition) archivable);
      } else {
        // archivable instanceof ProcessFullInstance because of previous checks
        recorder.remove((ProcessFullInstance) archivable);
      }
    } catch (Exception e) {
      throw new AtomicArchiveException("Can't remove archivable: "
          + archivable + " from history: " + recorder, e);
    }
  }

  /**
   * @param deploymentIdString
   * @return
   */
/*  public static int deploymentIdValueFrom(String deploymentIdString) {
    return Integer.parseInt(deploymentIdString.substring(DEPLOYMENT_IDPREFIX.length()));
  }

  public static String getNodeAsString(org.w3c.dom.Node n) throws TransformerException {
    Source source = new DOMSource(n);
    StringWriter stringWriter = new StringWriter();
    Result res = new StreamResult(stringWriter);
    TransformerFactory fact = TransformerFactory.newInstance();
    Transformer transformer = fact.newTransformer();
    transformer.transform(source, res);
    String resultAsString = stringWriter.getBuffer().toString();
    String prolog = "<?xml";
    if (resultAsString.startsWith(prolog)) {
      String prologEnd = "?>";
      int prologEndIndex = resultAsString.indexOf(prologEnd);
      return resultAsString.substring(prologEndIndex + prologEnd.length());
    }
    return resultAsString;
  } */

  public static String getStartingNodeName(String processId) {
    return "Process(" + processId + ")_startNode";
  }

  public static String getEndingNodeName(String processId) {
    return "Process(" + processId + ")_endNode";
  }

  public static Set<byte[]> getClasses(final Class< ? >... classes) {
    final Set<byte[]> classesSet = new HashSet<byte[]>();
    if (classes != null) {
      for (Class< ? > clazz : classes) {
        if (clazz == null) {
          throw new BonitaRuntimeException("Try to add a null class");
        }
        classesSet.add(getClassData(clazz));
      }
    }
    return classesSet;
  }

  public static byte[] getClassData(final Class< ? > clazz) {
    InputStream inputStream = null;
    try {
      final String className = clazz.getName();
      final String newClassName = className.substring(className.lastIndexOf(".") + 1) + ".class";
      inputStream = clazz.getResourceAsStream(newClassName);
      return Misc.getAllContentFrom(inputStream);
    } catch (Exception e) {
      throw new BonitaRuntimeException(e);
    } finally {
      Misc.close(inputStream);
    }
  }

  public static Set<String> copy(Set<String> src) {
    if (src == null) {
      return new HashSet<String>();
    }
    return new HashSet<String>(src);
  }

  public static List<String> copy(List<String> src) {
    if (src == null) {
      return new ArrayList<String>();
    }
    return new ArrayList<String>(src);
  }

  public static Map<String, String> copy(Map<String, String> src) {
    if (src == null) {
      return new HashMap<String, String>();
    }
    return new HashMap<String, String>(src);
  }

  public static Date copy(Date src) {
    if (src == null) {
      return null;
    }
    return (Date) src.clone();
  }

  public static Map<String, Object> getVariableValues(Map<String, Variable> variableMap) {
    Map<String, Object> res = new HashMap<String, Object>();
    if (variableMap != null) {
      for (Variable var : variableMap.values()) {
        res.put(var.getKey(), var.getValue());
      }
    }
    return res;
  }

  public static Map<String, Variable> getCopy(Map<String, Variable> variableMap) {
    return createVariableMap(getVariableValues(variableMap));
  }

  public static Map<String, Variable> createVariableMap(final Map<String, Object> variables) {
    if (variables == null || variables.isEmpty()) {
      return null;
    }
    Map<String, Variable> variableMap = new HashMap<String, Variable>();
    for (Map.Entry<String, Object> e : variables.entrySet()) {
      Object value = e.getValue();
      variableMap.put(e.getKey(), createVariable(e.getKey(), value));
    }
    return variableMap;
  }

  /**
   * Copied from the pvm (in VariableMap)
   * TODO: factorize code
   */
  public static Variable createVariable(String key, Object value) {
    Type type = null;
    Environment environment = Environment.getCurrent();
    if (environment != null) {
      VariableTypeResolver variableTypeResolver = environment.get(VariableTypeResolver.class);
      if (variableTypeResolver != null) {
        type = variableTypeResolver.findTypeByMatch(key, value);
      }
    }
    Variable variable = null;
    if (type != null) {
      Class< ? > variableClass = type.getVariableClass();
      try {
        variable = (Variable) variableClass.newInstance();
      } catch (Exception e) {
        throw new PvmException("couldn't instantiate variable instance class '" + variableClass.getName() + "'");
      }
      Converter converter = type.getConverter();
      variable.setConverter(converter);

    } else {
      if (value == null) {
        variable = new NullVariable();
      } else {
        variable = new UnpersistableVariable();
      }
    }

    variable.setKey(key);
    variable.setValue(getVariableValueCopy(value));
    return variable;
  }
  
  public static Object getVariableValueCopy(Object value) {
    if (value instanceof Enumeration) {
      Enumeration enumeration = (Enumeration) value;
      return new Enumeration(enumeration.getPossibleValues(), enumeration.getSelectedValue());
    } else if (value instanceof String) {
      return value;
    } else {
      // TODO: throw exception ?
      return value;
    }
  }

//  public static void printThreadStack(Thread thread) {
//    System.out.println("Printing Stacktrace of thread: " + thread);
//    StackTraceElement[] elts = thread.getStackTrace();
//    if (elts != null) {
//      for (StackTraceElement element : elts) {
//        System.out.println("  at " + element.toString());
//      }
//    }
//  }

  public static MyVisitor visitClass(final byte[] data) {
    final MyVisitor mv = new MyVisitor();
    final ClassReader cr = new ClassReader(data);
    cr.accept(mv, 1);
    return mv;
  }

}
